from datetime import datetime

import pytest

import preheat_open as po
import preheat_open.api as papi
from preheat_open.api.adapter import LocationNotFoundError
from preheat_open.loadable_types import (
    ComfortProfile,
    ComfortProfiles,
    ElectricityPrices,
)

TEST_LOCATION_ID = 2756


@pytest.fixture
def date_range():
    return po.DateRange(
        start=datetime(2024, 1, 1),
        end=datetime(2024, 1, 15),
        resolution=po.TimeResolution.HOUR,
    )


@pytest.fixture
def real_location(api_adapter):
    return api_adapter.get_location(location_id=TEST_LOCATION_ID)


@pytest.fixture
def real_unit(real_location):
    return real_location.units[0]


def test_get_features(api_adapter, real_location):
    features = api_adapter.get_features(location_id=real_location.id)
    assert isinstance(features, list)


def test_get_all_locations_information(api_adapter):
    locations_info = api_adapter.get_all_locations_information()
    assert isinstance(locations_info, list)
    assert all(isinstance(info, po.LocationInformation) for info in locations_info)


def test_get_all_locations_collection(api_adapter):
    collection = api_adapter.get_all_locations_collection()
    assert isinstance(collection, po.Collection)


def test_get_location(real_location):
    assert isinstance(real_location, po.Location)


def test_get_location_not_found(api_adapter):
    with pytest.raises(LocationNotFoundError):
        api_adapter.get_location(location_id=999)


def test_get_locations(api_adapter):
    locations = api_adapter.get_locations(location_ids=[TEST_LOCATION_ID])
    assert isinstance(locations, list)
    assert all(isinstance(loc, po.Location) for loc in locations)


def test_get_devices(api_adapter):
    devices = api_adapter.get_devices(location_id=TEST_LOCATION_ID)
    assert isinstance(devices, list)
    assert all(isinstance(device, po.Device) for device in devices)


def test_load_measurements(api_adapter, date_range):
    components = [
        po.Component(
            id=1,
            parent=po.Unit(
                id=1,
                type=papi.UnitType.WEATHER_FORECAST,
                location=po.Location(id=TEST_LOCATION_ID),
            ),
        )
    ]
    api_adapter.load_measurements(components=components, date_range=date_range)
    assert components[0].measurements[date_range.resolution] is not None


def test_get_price_components(api_adapter, real_location):
    test_supply_point = real_location.supply_points[0]
    price_components = api_adapter.get_price_components(
        supply_point_ids=[test_supply_point.id]
    )
    assert isinstance(price_components, dict)
    assert all(isinstance(key, int) for key in price_components.keys())
    assert all(isinstance(value, list) for value in price_components.values())
    assert all(
        isinstance(item, po.AppliedPriceComponent)
        for sublist in price_components.values()
        for item in sublist
    )


def test_get_price_data(api_adapter, date_range):
    price_data = api_adapter.get_price_data(
        date_range=date_range, price_component_ids=[1]
    )
    assert isinstance(price_data, dict)
    assert all(isinstance(key, int) for key in price_data.keys())
    assert all(isinstance(value, po.PriceData) for value in price_data.values())


def test_get_setpoint_schedule(api_adapter, real_location, date_range):
    control_unit = list(real_location.get_units(type="controls"))[0]
    setpoint_schedule = api_adapter.get_setpoint_schedule(
        control_unit=control_unit, date_range=date_range
    )
    assert isinstance(setpoint_schedule, po.SetpointSchedule)


def test_get_comfort_profile_setpoints(api_adapter, date_range, real_location):
    comfort_profiles = api_adapter.get_comfort_profile_setpoints(
        date_range=date_range, location=real_location
    )
    assert isinstance(comfort_profiles, ComfortProfiles)
    assert all(
        isinstance(profile, ComfortProfile)
        for id, profile in comfort_profiles.profiles.items()
    )


def test_get_electricity_prices(api_adapter, date_range, real_location):
    electricity_unit = list(real_location.get_units(type="electricity"))[0]
    electricity_prices = api_adapter.get_electricity_prices(
        unit=electricity_unit, date_range=date_range
    )
    assert isinstance(electricity_prices, ElectricityPrices)



# pylint: disable=E1101
from __future__ import annotations

import logging
import os
import re
import sys
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Generator, Generic, Optional, Type, TypeVar, Union

import yaml

logger = logging.getLogger(__name__)

_USUAL_CONFIG_PATHS = [
    "~/.preheat",
    "/etc/opt/preheat",
    "/etc/preheat",
]

_ENV_PREFIX = "PREHEAT_"


def _env_var_name(obj: ConfigBase | Type[ConfigBase], parameter: str) -> str:
    """Generate environment variable name for a given parameter.

    Args:
        obj (ConfigBase): The configuration object
        parameter (str): The parameter name

    Returns:
        str: The environment variable name
    """
    class_name = obj.__name__ if isinstance(obj, type) else obj.__class__.__name__
    if parameter == "runtime_mode":
        return f"{_ENV_PREFIX}RUNTIME_MODE"
    return f"{_ENV_PREFIX}{class_name.upper()}_{parameter.upper()}"


class UpdateProxy:
    """Proxy object that allows updating environment variables for configuration properties."""

    def __init__(self, config_obj):
        self._config_obj: ConfigBase = config_obj

    def all(self) -> None:
        """Update all environment variables to match the current configuration object."""
        for key, value in self._config_obj.iter_items():
            if not key.startswith("_") and not callable(getattr(self._config_obj, key)):
                os.environ[_env_var_name(self._config_obj, key)] = str(value)

    def __setattr__(self, name: str, value: Any) -> None:
        if name == "_config_obj":
            # Allow setting the internal reference
            super().__setattr__(name, value)
            return

        if (
            hasattr(self._config_obj, name)
            and not name.startswith("_")
            and not callable(getattr(self._config_obj, name))
        ):
            # Update environment variable
            os.environ[_env_var_name(self._config_obj, name)] = str(value)
            # Update the actual attribute
            setattr(self._config_obj, name, value)
        else:
            raise AttributeError(
                f"{self._config_obj.__class__.__name__} has no attribute '{name}'"
            )


T = TypeVar("T", bound="ConfigBase")


class ConfigBase:
    """Base class for all configuration objects.

    This class provides common functionality for all configuration classes,
    including environment variable access and sensitive data handling.

    Attributes:
        runtime_mode (int): The current runtime mode
    """

    runtime_mode: int = 10
    _sensitive_fields: set[str] = set()

    def __init__(self):
        """Initialize configuration from environment variables."""
        self._update_proxy = UpdateProxy(self)
        self._load_from_env()

    @property
    def update(self) -> UpdateProxy:
        """Get a proxy object to update configuration attributes and environment variables."""
        return self._update_proxy

    def iter_items(self) -> Generator[tuple[str, Any], None, None]:
        """Iterate over all attributes of the configuration object.

        Yields:
            tuple[str, Any]: A tuple of attribute name and value
        """
        for key in dir(self):
            if (
                not key.startswith("_")
                and not callable(getattr(self, key))
                and key != "update"
            ):
                yield key, getattr(self, key)

    class TempContext(Generic[T]):
        """Context manager for temporarily setting environment variables."""

        def __init__(self, config_obj: T):
            self.config_obj = config_obj
            self.original_env = {}

        def __enter__(self) -> T:
            # Save original environment variable values
            for key, value in self.config_obj.iter_items():
                env_var = _env_var_name(self.config_obj, key)
                self.original_env[env_var] = os.environ.get(env_var)

            # Update environment variables with current config values
            self.config_obj.update.all()
            return self.config_obj

        def __exit__(self, exc_type, exc_val, exc_tb):
            # Restore original environment variables
            for env_var, value in self.original_env.items():
                if value is None:
                    if env_var in os.environ:
                        del os.environ[env_var]
                else:
                    os.environ[env_var] = value

    def temporary(self: T) -> TempContext[T]:
        """Return a context manager that temporarily sets environment variables.

        Usage:
            with config.temporary():
                # Environment variables reflect config values here
            # Original environment variables restored here

        Returns:
            TempContext: A context manager that preserves the specific config type
        """
        return self.TempContext(self)

    def _load_from_env(self):
        """Load configuration values from environment variables."""

        for key, value in self.iter_items():
            type_ = type(value)

            env_value = os.environ.get(_env_var_name(self, key))
            if env_value is not None:
                setattr(self, key, type_(env_value))

        logger.debug("Loaded configuration: %s", self.pretty_print())

    def __str__(self) -> str:
        return self.pretty_print()

    def pretty_print(self, anonymize: bool | None = None) -> str:
        """Create a string representation with sensitive fields masked.

        Returns:
            str: String representation of the configuration
        """
        anonymize = (
            self.runtime_mode != RuntimeMode.DEVELOPMENT.value
            if anonymize is None
            else anonymize
        )
        attrs = [
            f"{key}={'*****' if key in self._sensitive_fields and anonymize else value}"
            for key, value in self.iter_items()
        ]

        return f"{self.__class__.__name__}({', '.join(attrs)})"


class PersonalConfig(ConfigBase):
    """Personal configuration settings.

    Holds personal identification and preferences information.

    Attributes:
        name (str): The user's name
        email (str): The user's email address
        timezone (str): The user's timezone
        runtime_mode (int): The current runtime mode
    """

    name: str = ""
    email: str = ""
    timezone: str = ""
    _sensitive_fields: set[str] = {"name", "email"}


class LoggingConfig(ConfigBase):
    """Logging configuration settings.

    Controls the application's logging behavior.

    Attributes:
        level (int): The logging level (corresponds to standard logging levels)
        format (str): The format string for log messages
        auto_setup (bool): Whether to automatically set up logging on initialization
        directory (str): Path to the log directory
        library_name (str): Name of the library for logger namespace
        runtime_mode (int): The current runtime mode
    """

    level: int = logging.INFO
    format: str = "%(asctime)-23s  %(levelname)-8s  %(name)-32s  %(message)-80s"
    auto_setup: bool = False
    directory: str = "~/.preheat/.logs"

    def __init__(self):
        super().__init__()
        if self.auto_setup:
            self.setup_logging()

    def setup_logging(self) -> None:
        """Set up logging based on the configuration."""
        # Automatically determine the library name from the module structure
        library_name = self._get_library_name()

        # Create log directory if it doesn't exist
        log_dir = Path(self.directory).expanduser()
        log_dir.mkdir(parents=True, exist_ok=True)

        # Get timestamp for log file
        today = datetime.now()
        file_name = f"{library_name}-{today.strftime('%Y-%m-%d %H:%M:%S')}.log"

        # Reset only loggers that belong to our application
        for logger_name in list(logging.root.manager.loggerDict.keys()):
            if logger_name.startswith(library_name):
                logger_instance = logging.getLogger(logger_name)
                # Save the propagate setting to restore it later
                propagate = logger_instance.propagate
                # Remove handlers
                for handler in logger_instance.handlers[:]:
                    logger_instance.removeHandler(handler)
                    handler.close()
                # Set level but don't change propagate setting
                logger_instance.setLevel(self.level)
                logger_instance.propagate = propagate

        # Create handlers
        file_handler = logging.FileHandler(log_dir / file_name)
        stream_handler = logging.StreamHandler()

        # Create formatter
        formatter = logging.Formatter(self.format)
        file_handler.setFormatter(formatter)
        stream_handler.setFormatter(formatter)

        # Configure the root logger for our application's modules
        lib_logger = logging.getLogger(library_name)
        lib_logger.setLevel(self.level)
        lib_logger.handlers = []  # Clear any existing handlers
        lib_logger.addHandler(file_handler)
        lib_logger.addHandler(stream_handler)
        lib_logger.propagate = False  # Don't propagate to root logger

        logger.debug(f"Logging configured for {library_name} with level: {self.level}")

    def _get_library_name(self) -> str:
        """Automatically determine the library name from the module structure.

        Returns:
            str: The top-level package name
        """
        # Get the module name of this file (e.g., 'preheat_open.api.configuration')
        module_name = __name__

        # Extract the top-level package name (e.g., 'preheat_open')
        parts = module_name.split(".")
        if len(parts) > 0:
            return parts[0]

        # Fallback to a default name if we can't determine it
        return "preheat_open"


class CacheConfig(ConfigBase):
    """Cache configuration settings.

    Controls how the application caches data.

    Attributes:
        type (str): The cache type to use
        directory (str): Path to the cache directory
        size_limit (float): Maximum size of the cache in bytes
        time_to_live (int): Maximum lifetime of cached items in seconds
        runtime_mode (int): The current runtime mode
    """

    type: str = ""
    directory: str = "~/.preheat/.cache"
    size_limit: float = 5e8  # 500 MB
    time_to_live: int = 60 * 60 * 24  # 1 day


class NeogridApiConfig(ConfigBase):
    """Neogrid API configuration settings.

    Contains credentials and settings for accessing the Neogrid API.

    Attributes:
        token (str): API access token
        url (str): Base URL for API requests
        runtime_mode (int): The current runtime mode
    """

    token: str = "KVkIdWLKac5XFLCs2loKb7GUitkTL4uJXoSyUFIZkVgWuCk8Uj"
    url: str = "https://api.neogrid.dk/public/api/v1"
    _sensitive_fields: set[str] = {"token"}


class CaseConverter:
    """Convert a string to a case-insensitive version.

    This class is used to convert configuration keys to a specific case
    for easier access and comparison.

    Attributes:
        value (str): The original string value
    """

    def __init__(self):
        self.pattern = re.compile(r"(?<=[a-z])(?=[A-Z])|(?<=[A-Z])(?=[A-Z][a-z])")

    def snake_to_pascal(self, value: str) -> str:
        """Convert a snake_case string to PascalCase.

        Args:
            value (str): The snake_case string

        Returns:
            str: The converted PascalCase string
        """
        return "".join(word.capitalize() for word in value.split("_"))

    def pascal_to_snake(self, value: str) -> str:
        """Convert a PascalCase string to snake_case.

        Args:
            value (str): The PascalCase string

        Returns:
            str: The converted snake_case string
        """
        return self.pattern.sub("_", value).lower()


class ConfigLoader:
    """Loads configuration from YAML and sets environment variables.

    Responsible for loading configuration data from a YAML file and setting
    the corresponding environment variables that will be used by configuration classes.

    This loader automatically adapts to any ConfigBase subclasses in the system.
    """

    def __init__(
        self,
        data: dict | None = None,
        path: Optional[str] = None,
        force_reload: bool = True,
    ):
        """Initialize the config loader.

        Args:
            default_config_path (str): Path to the default configuration file
        """
        self.path = path or self.find_config_path()
        self.data = data or {}
        self._config_classes: list[Type] = ConfigBase.__subclasses__()
        if not self.data and self.path:
            self.load_config_file(path=self.path)

        self._set_environment_variables(force_reload=force_reload)

    def find_config_path(self) -> str | None:
        """Find the first available configuration directory and its config file.

        Searches first in private paths, then falls back to public paths.

        Returns:
            str: Path to the configuration file
        """
        for path in _USUAL_CONFIG_PATHS:
            expanded_path = Path(path).expanduser() / "config.yaml"
            if expanded_path.exists():
                return str(expanded_path)
        return None

    def load_config_file(self, path: Optional[Union[str, Path]] = None) -> None:
        """Load configuration from a YAML file.

        Args:
            file_path (Optional[Union[str, Path]]): Path to configuration file.
                If None, the default path will be used.

        Raises:
            FileNotFoundError: If the configuration file doesn't exist
        """
        path = path or self.path

        if path is None:
            raise FileNotFoundError("No configuration file path specified or found")

        path_obj = Path(path).expanduser().resolve()
        if not path_obj.exists():
            raise FileNotFoundError(f"Configuration file not found: {path_obj}")

        with open(path_obj, "r") as f:
            self.data = yaml.safe_load(f)

    def _set_environment_variables(self, force_reload: bool = False) -> None:
        """Set environment variables from loaded configuration."""
        if not self.data:
            return

        # Iterate through each config class
        for config_class in self._config_classes:
            # Determine section name from class name
            section_name = CaseConverter().pascal_to_snake(
                config_class.__name__.removesuffix("Config")
            )

            # Check if this section exists in the config data
            if section_name in self.data and isinstance(self.data[section_name], dict):
                section_data = self.data[section_name]

                # Set environment variables for each key in this section
                for key, value in section_data.items():
                    if value is not None:
                        env_var_name = _env_var_name(config_class, key)
                        if force_reload or env_var_name not in os.environ:
                            os.environ[env_var_name] = str(value)


if not any(
    caller.endswith("import ConfigLoader")
    for caller in sys._getframe(1).f_code.co_names
):
    ConfigLoader(force_reload=False)


class RuntimeMode(Enum):
    """
    Enum representing the runtime mode of the application.

    Attributes:
        PRODUCTION (int): Production mode.
        DEVELOPMENT (int): Development mode.
        TEST (int): Test mode.
    """

    PRODUCTION = 0
    DEVELOPMENT = 10
    TEST = 20

    def __eq__(self, other: object) -> bool:
        """
        Check equality with another object.

        :param other: The object to compare with.
        :type other: object
        :return: True if equal, False otherwise.
        :rtype: bool
        """
        if isinstance(other, RuntimeMode):
            return self.value == other.value
        elif isinstance(other, int):
            return self.value == other
        return False


def run_if_production_mode(
    function: Callable, runtime_mode: int | None = None, *args, **kwargs
) -> Any:
    """
    Runs a function if the application is in production mode.

    :param configuration: The configuration settings for the application.
    :type configuration: Configuration
    :param function: The function to run.
    :type function: callable
    """
    runtime_mode = (
        runtime_mode if runtime_mode is not None else PersonalConfig().runtime_mode
    )

    if runtime_mode == RuntimeMode.PRODUCTION.value:
        return function(*args, **kwargs)
    else:
        log = f"Function not run because application is not in production mode. Function: {function}({args}, {kwargs})"
        logger.warning(log[:1000])


class RuntimeModeError(Exception):
    """
    Exception raised for errors in the runtime mode.
    """

    pass



import logging
from dataclasses import dataclass, field
from typing import Callable, Generator

import pandas as pd

from .interfaces import Adapter
from .location import Location, LocationInformation
from .measurements import get_measurements
from .query import Query
from .time import ZoneInfo
from .unit import Component, Device, Unit
from .zone import Zone

logger = logging.getLogger(__name__)


@dataclass
class Collection:
    """
    Represents a collection of locations.

    :ivar locations: The locations in the collection.
    """

    locations: dict[LocationInformation, Location] = field(default_factory=dict)
    adapter: Adapter = None
    _models_loaded: bool = field(default=False)
    """
    The locations in the collection.
    """

    def __post_init__(self) -> None:
        """
        Post-initialization processing to set up the collection.
        """
        if all([loc is not None for loc in self.locations.values()]):
            self._models_loaded = True

    @classmethod
    def from_location_ids(
        cls, location_ids: list[int], adapter: Adapter, **kwargs
    ) -> "Collection":
        """
        Creates a collection from a list of locations.

        :param location_list: The list of locations.
        :return: The collection.
        """
        locations = adapter.get_locations(location_ids)
        return cls(
            locations={l.information: l for l in locations},
            adapter=adapter,
            **kwargs,
        )

    @classmethod
    def from_location_information_list(
        cls, location_information_list: list[LocationInformation], **kwargs
    ) -> "Collection":
        """
        Creates a collection from a list of location information.

        :param location_information_list: The list of location information.
        :return: The collection.
        """
        return cls(
            locations={loc_info: None for loc_info in location_information_list},
            **kwargs,
        )

    def get_keys(self, **kwargs) -> Generator[None, None, LocationInformation]:
        for loc_info in self.locations.keys():
            if all(getattr(loc_info, key) == value for key, value in kwargs.items()):
                yield loc_info

    def load_building_models(self) -> None:
        """
        Loads the building models for each location in the collection.
        """
        ids_to_load = [
            loc_info.id for loc_info, model in self.locations.items() if model is None
        ]
        if ids_to_load:
            models = self.adapter.get_locations(ids_to_load)
            for model in models:
                key = next(self.get_keys(id=model.id))
                self.locations[key] = model

    def filter(self, location_info_filter: Callable) -> "Collection":
        """
        Filters the collection based on a filter function.

        :param filter_func: The filter function.
        :return: The filtered collection.
        """
        return Collection(
            locations={
                loc_info: loc
                for loc_info, loc in self.locations.items()
                if location_info_filter(loc_info)
            },
            adapter=self.adapter,
        )

    def get_measurements(
        self,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Retrieves measurements for the location.

        :param components: The components to get measurements for.
        :type components: list[Component] | Query | dict
        :param date_range: The date range for the measurements.
        :type date_range: DateRange, optional
        :param mapper: The mapper to apply to the measurements.
        :type mapper: MapApplier, optional
        :param kwargs: Additional arguments.
        :return: A DataFrame containing the measurements.
        :rtype: pd.DataFrame
        """
        if not self._models_loaded:
            self.load_building_models()
        return get_measurements(
            obj=self,
            **kwargs,
        )

    def get_units(
        self,
        query: Query | None = None,
        **kwargs,
    ) -> Generator[Unit, None, None]:
        """
        Retrieves units associated with the location.

        :param query: The query to filter units.
        :type query: Query, optional
        :param kwargs: Additional arguments.
        :return: A generator yielding units.
        :rtype: Generator[Unit, None, None]
        """
        if not self._models_loaded:
            self.load_building_models()
        for location in self.locations.values():
            yield from location.get_units(
                query=query,
                **kwargs,
            )

    def get_devices(
        self,
        query: Query | list[Query] | None = None,
        **kwargs,
    ) -> Generator[Device, None, None]:
        """
        Retrieves devices associated with the location.

        :param query: The query to filter devices.
        :type query: Query, optional
        :param kwargs: Additional arguments.
        :return: A generator yielding devices.
        :rtype: Generator[Device, None, None]
        """
        if not self._models_loaded:
            self.load_building_models()
        for location in self.locations.values():
            yield from location.get_devices(
                query=query,
                **kwargs,
            )

    def get_zones(
        self,
        query: Query | list[Query] | None = None,
        **kwargs,
    ) -> Generator[Zone, None, None]:
        """
        Retrieves zones associated with the location.

        :param query: The query to filter zones.
        :type query: Query, optional
        :param kwargs: Additional arguments.
        :return: A generator yielding zones.
        :rtype: Generator[Zone, None, None]
        """
        if not self._models_loaded:
            self.load_building_models()
        for location in self.locations.values():
            yield from location.get_zones(
                query=query,
                **kwargs,
            )

    def get_components(
        self,
        query: Query | list[Query] | None = None,
        **kwargs,
    ) -> Generator[Component, None, None]:
        """
        Retrieves components associated with the location.

        :param query: The query to filter components.
        :type query: Query | list[Query], optional
        :param kwargs: Additional arguments.
        :return: A generator yielding components.
        :rtype: Generator[Component, None, None]
        """
        if not self._models_loaded:
            self.load_building_models()
        for location in self.locations.values():
            yield from location.get_components(
                query=query,
                **kwargs,
            )

    @property
    def timezone(self) -> str:
        """
        The timezone of the location.

        :return: The timezone.
        :rtype: str
        """
        if hasattr(self.adapter, "configuration"):
            timezone = getattr(self.adapter.configuration.personal, "timezone")
            logger.debug(
                f"Tried using timezone on {self.__class__.__name__}. Fallback to timezone from configuration: {timezone}"
            )
        else:
            timezone = ZoneInfo("Europe/Copenhagen")
            logger.debug(
                f"Tried using timezone on {self.__class__.__name__}. Fallback to default timezone: {timezone}"
            )
        return timezone



import pathlib
from typing import Generator

import pytest

import preheat_open.api as papi
import preheat_open.api.mocks as mocks


@pytest.fixture(scope="session", autouse=True)
def test_config() -> None:
    from preheat_open.api.configuration import ConfigLoader

    file_path = f"{str(pathlib.Path(__file__).parent.resolve())}/test/mocks/configurations/config.yaml"
    ConfigLoader(path=file_path, force_reload=True)


@pytest.fixture(scope="session")
def mock_adapter() -> Generator[mocks.MockApiAdapter, None, None]:
    with mocks.MockApiAdapter(
        mocks_path=f"{str(pathlib.Path(__file__).parent.resolve())}/test/mocks",
    ) as adapter:
        yield adapter


@pytest.fixture(scope="session")
def api_adapter() -> Generator[papi.ApiAdapter, None, None]:
    with papi.ApiAdapter() as adapter:
        yield adapter



"""
The preheat_open package provides methods to load building data from `Neogrids <https://www.neogrid.dk/>`_ platform
"""

# Imports for backwards compatibility of dependent code

import os
import re
import sys

from preheat_open.api import *

from . import _version
from .collection import Collection
from .interfaces import Adapter, BuildingModel
from .loadable_types import (
    ComfortProfile,
    ComfortProfiles,
    ElectricityPriceItem,
    ElectricityPrices,
    LoadableData,
    OperationType,
    ScheduleItem,
    Setpoint,
    SetpointSchedule,
)
from .location import Location, LocationCharacteristics, LocationInformation
from .measurements import AttrMap, Mapper, MeasurementRequest, PostProcess, StrMap
from .query import unique
from .supplypoint import AppliedPriceComponent, PriceData, SupplyPoint
from .time import DateRange, TimeResolution
from .unit import Component, Device, Unit, UnitDescriptors
from .zone import Zone

__version__ = _version.get_versions()["version"]



from .adapter import ApiAdapter, TimestampType, building_model
from .cache import ShelveCache, SimplePickleCache
from .types import (
    ComponentQuery,
    ComponentType,
    DeviceQuery,
    RelatedUnitType,
    UnitDescriptorType,
    UnitQuery,
    UnitSubType,
    UnitType,
    ZoneQuery,
)



import os
import tempfile
import unittest
import unittest.mock
from pathlib import Path

import pytest
import yaml

from preheat_open.api.configuration import (
    CacheConfig,
    CaseConverter,
    ConfigBase,
    ConfigLoader,
    LoggingConfig,
    NeogridApiConfig,
    PersonalConfig,
    RuntimeMode,
    RuntimeModeError,
    UpdateProxy,
    _env_var_name,
    run_if_production_mode,
)

# === Fixtures ===


@pytest.fixture
def clean_env():
    """Fixture to provide a clean environment for tests."""
    old_environ = os.environ.copy()
    os.environ.clear()
    # Ensure runtime mode is set to TEST
    os.environ["PREHEAT_RUNTIME_MODE"] = str(RuntimeMode.TEST.value)
    yield
    os.environ.clear()
    os.environ.update(old_environ)


@pytest.fixture
def temp_config_file():
    """Create a temporary config file for testing."""
    with tempfile.NamedTemporaryFile(suffix=".yaml", delete=False, mode="w") as tmp:
        test_config = {
            "personal": {
                "name": "Test User",
                "email": "test@example.com",
                "timezone": "UTC",
            },
            "logging": {"level": 20, "auto_setup": False},  # INFO
            "neogrid_api": {
                "token": "test-token",
                "url": "https://test-api.example.com",
            },
        }
        yaml.dump(test_config, tmp)

    yield tmp.name

    # Clean up
    os.unlink(tmp.name)


@pytest.fixture
def test_config_class():
    """Create a test configuration class."""

    class TestConfig(ConfigBase):
        test_string: str = "default"
        test_int: int = 42
        test_float: float = 3.14
        _sensitive_fields: set[str] = {"test_string"}

    return TestConfig


# === Tests for ConfigBase and its methods ===


def test_config_base_init(clean_env, test_config_class):
    """Test initialization of a ConfigBase class."""
    os.environ["PREHEAT_TESTCONFIG_TEST_STRING"] = "from_env"
    os.environ["PREHEAT_TESTCONFIG_TEST_INT"] = "100"

    config = test_config_class()

    assert config.test_string == "from_env"
    assert config.test_int == 100
    assert config.test_float == 3.14  # unchanged


def test_update_proxy(clean_env, test_config_class):
    """Test UpdateProxy for modifying config and environment variables."""
    config = test_config_class()

    # Test setting a value through the proxy
    config.update.test_string = "updated"

    # Check that both config and environment were updated
    assert config.test_string == "updated"
    assert os.environ["PREHEAT_TESTCONFIG_TEST_STRING"] == "updated"

    # Test the .all() method
    config.test_int = 999  # Directly set attribute (env not updated)
    config.update.all()
    assert os.environ["PREHEAT_TESTCONFIG_TEST_INT"] == "999"

    # Test error on non-existent attribute
    with pytest.raises(AttributeError):
        config.update.nonexistent = "value"


def test_temporary_context(clean_env, test_config_class):
    """Test the temporary context manager for environment variables."""
    # Set initial environment
    os.environ["PREHEAT_TESTCONFIG_TEST_STRING"] = "initial"

    config = test_config_class()
    assert config.test_string == "initial"

    # Modify the config but not the environment
    config.test_string = "modified"
    assert config.test_string == "modified"
    assert os.environ["PREHEAT_TESTCONFIG_TEST_STRING"] == "initial"

    # Use the temporary context
    with config.temporary():
        assert os.environ["PREHEAT_TESTCONFIG_TEST_STRING"] == "modified"
        config.update.test_string = "in_context"
        assert os.environ["PREHEAT_TESTCONFIG_TEST_STRING"] == "in_context"

    # Environment should be restored after context
    assert os.environ["PREHEAT_TESTCONFIG_TEST_STRING"] == "initial"
    # But config object should keep its value
    assert config.test_string == "in_context"


def test_pretty_print(clean_env, test_config_class):
    """Test pretty printing of configuration with/without anonymization."""
    config = test_config_class()
    config.test_string = "sensitive data"

    # Test anonymized printing (default in non-development mode)
    assert "*****" in config.pretty_print()
    assert "sensitive data" not in config.pretty_print()

    # Test non-anonymized printing
    assert "sensitive data" in config.pretty_print(anonymize=False)

    # Test str representation
    assert str(config) == config.pretty_print()


def test_iter_items(clean_env, test_config_class):
    """Test iteration over configuration items."""
    config = test_config_class()

    items = dict(config.iter_items())
    assert "test_string" in items
    assert "test_int" in items
    assert "test_float" in items
    assert "_sensitive_fields" not in items  # Private attributes excluded
    assert "update" not in items  # Properties excluded
    assert "iter_items" not in items  # Methods excluded


# === Tests for specific configuration subclasses ===


def test_personal_config(clean_env):
    """Test PersonalConfig class."""
    os.environ["PREHEAT_PERSONALCONFIG_NAME"] = "John Doe"
    os.environ["PREHEAT_PERSONALCONFIG_EMAIL"] = "john@example.com"

    config = PersonalConfig()

    assert config.name == "John Doe"
    assert config.email == "john@example.com"
    assert "name" in config._sensitive_fields
    assert "email" in config._sensitive_fields


def test_logging_config(clean_env, monkeypatch):
    """Test LoggingConfig class."""
    # Mock the actual logging setup to avoid side effects
    setup_mock = unittest.mock.Mock()
    monkeypatch.setattr(LoggingConfig, "setup_logging", setup_mock)

    # Test without auto_setup
    config = LoggingConfig()
    assert setup_mock.call_count == 0

    # Test with auto_setup
    os.environ["PREHEAT_LOGGINGCONFIG_AUTO_SETUP"] = "True"
    config = LoggingConfig()
    assert setup_mock.call_count == 1


def test_cache_config(clean_env):
    """Test CacheConfig class."""
    os.environ["PREHEAT_CACHECONFIG_SIZE_LIMIT"] = "1000000"
    os.environ["PREHEAT_CACHECONFIG_TIME_TO_LIVE"] = "3600"

    config = CacheConfig()

    assert config.size_limit == 1000000.0
    assert config.time_to_live == 3600


def test_neogrid_api_config(clean_env):
    """Test NeogridApiConfig class."""
    os.environ["PREHEAT_NEOGRIDAPICONFIG_TOKEN"] = "secret-token"
    os.environ["PREHEAT_NEOGRIDAPICONFIG_URL"] = "https://api.example.com"

    config = NeogridApiConfig()

    assert config.token == "secret-token"
    assert config.url == "https://api.example.com"
    assert "token" in config._sensitive_fields


# === Tests for ConfigLoader ===


def test_config_loader_with_file(clean_env, temp_config_file):
    """Test ConfigLoader with a file path."""
    loader = ConfigLoader(path=temp_config_file)

    # Check that environment variables were set correctly
    assert os.environ["PREHEAT_PERSONALCONFIG_NAME"] == "Test User"
    assert os.environ["PREHEAT_PERSONALCONFIG_EMAIL"] == "test@example.com"
    assert os.environ["PREHEAT_LOGGINGCONFIG_LEVEL"] == "20"
    assert os.environ["PREHEAT_NEOGRIDAPICONFIG_TOKEN"] == "test-token"


def test_config_loader_with_data(clean_env):
    """Test ConfigLoader with direct data."""
    test_data = {"personal": {"name": "Data User", "email": "data@example.com"}}

    loader = ConfigLoader(data=test_data)

    # Check that environment variables were set correctly
    assert os.environ["PREHEAT_PERSONALCONFIG_NAME"] == "Data User"
    assert os.environ["PREHEAT_PERSONALCONFIG_EMAIL"] == "data@example.com"


# === Tests for utility functions ===


def test_env_var_name():
    """Test environment variable name generation."""

    class TestConfig(ConfigBase):
        test_param: str = ""

    # Test with class
    assert _env_var_name(TestConfig, "test_param") == "PREHEAT_TESTCONFIG_TEST_PARAM"

    # Test with instance
    config = TestConfig()
    assert _env_var_name(config, "test_param") == "PREHEAT_TESTCONFIG_TEST_PARAM"

    # Test runtime_mode special case
    assert _env_var_name(config, "runtime_mode") == "PREHEAT_RUNTIME_MODE"


def test_case_converter():
    """Test the CaseConverter utility."""
    converter = CaseConverter()

    # Test snake to pascal
    assert converter.snake_to_pascal("test_case") == "TestCase"
    assert converter.snake_to_pascal("multiple_word_case") == "MultipleWordCase"

    # Test pascal to snake
    assert converter.pascal_to_snake("TestCase") == "test_case"
    assert converter.pascal_to_snake("MultipleWordCase") == "multiple_word_case"
    assert converter.pascal_to_snake("ABCTest") == "abc_test"


def test_runtime_mode():
    """Test RuntimeMode enum."""
    # Test equality
    assert RuntimeMode.PRODUCTION == RuntimeMode.PRODUCTION
    assert RuntimeMode.PRODUCTION == 0
    assert RuntimeMode.DEVELOPMENT == 10
    assert RuntimeMode.TEST == 20

    # Test inequality
    assert RuntimeMode.PRODUCTION != RuntimeMode.DEVELOPMENT
    assert RuntimeMode.DEVELOPMENT != RuntimeMode.TEST
    assert RuntimeMode.PRODUCTION != "production"


def test_run_if_production_mode():
    """Test run_if_production_mode function."""
    test_func = unittest.mock.Mock(return_value="executed")

    # Should run in production mode
    result = run_if_production_mode(
        function=test_func,
        runtime_mode=RuntimeMode.PRODUCTION.value,
        arg1="arg1",
        arg2="value",
    )
    assert result == "executed"
    test_func.assert_called_once_with(arg1="arg1", arg2="value")

    test_func.reset_mock()

    # Should not run in other modes
    result = run_if_production_mode(
        function=test_func, runtime_mode=RuntimeMode.DEVELOPMENT.value, arg1="arg1"
    )
    assert result is None
    test_func.assert_not_called()



from dataclasses import dataclass

import pytest

from preheat_open.interfaces import Adapter, BuildingModel, Factory, log_method_call


@dataclass
class TClass:
    attr1: int
    attr2: str = ""
    attr3: str = ""


class MockFactory(Factory):
    _return_class = TClass
    _translation_dict = {
        "attr1": "Attr1",
        "attr2": "Attr2",
        "attr3": "Attr3",
    }


@pytest.fixture(scope="module")
def test_input_dict():
    return {
        "Attr1": 23,
        "Attr2": "Foo",
    }


def test_build_factory(test_input_dict):
    obj = MockFactory(test_input_dict.copy()).build()
    assert isinstance(obj, TClass)
    assert obj.attr1 == 23
    assert obj.attr2 == "Foo"
    assert obj.attr3 == ""


def test_build_factory_with_params(test_input_dict):
    obj2 = MockFactory(test_input_dict.copy()).build(Attr3="Bar")
    assert isinstance(obj2, TClass)
    assert obj2.attr1 == 23
    assert obj2.attr2 == "Foo"
    assert obj2.attr3 == "Bar"


def test_factory_translate():
    d = MockFactory.translate(TClass(23, "Foo"))
    assert d == {"Attr1": 23, "Attr2": "Foo", "Attr3": ""}



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, ClassVar, TypeVar, final

from requests.models import Response

from .loadable_types import ComfortProfile, ElectricityPrices
from .time import DateRange

if TYPE_CHECKING:
    from .collection import Collection
    from .loadable_types import SetpointSchedule
    from .location import Location, LocationInformation
    from .supplypoint import AppliedPriceComponent, PriceData
    from .unit import Component, Device, Unit

logger = logging.getLogger(__name__)


def log_method_call(logger=logger, level="debug"):
    def decorator(method):
        @wraps(method)
        def wrapper(self, *args, **kwargs):
            # Format the arguments and keyword arguments
            args_str = ", ".join(repr(arg) for arg in args)
            kwargs_str = ", ".join(f"{k}={v!r}" for k, v in kwargs.items())
            all_args_str = ", ".join(filter(None, [args_str, kwargs_str]))

            # Log the formatted message
            getattr(logger, level)(
                "%s.%s(%s)",
                self.__class__.__name__,
                method.__name__,
                all_args_str,
            )
            return method(self, *args, **kwargs)

        return wrapper

    return decorator


@dataclass
class BuildingModel:
    component_query_type: type
    component_type: type
    unit_query_type: type
    unit_type: type
    unit_subtype: type
    device_query_type: type
    device_type: type
    zone_query_type: type
    zone_type: type
    control_setting_type: type


T = TypeVar("T", bound="Adapter")


@dataclass
class Adapter(ABC):
    building_model: ClassVar[BuildingModel]

    @abstractmethod
    def open(self) -> None:
        pass

    @abstractmethod
    def close(self) -> None:
        pass

    def __enter__(self: T) -> T:
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> bool:
        self.close()
        return False

    @abstractmethod
    def get_features(self, location_id: int):
        pass

    @abstractmethod
    def get_all_locations_information(self) -> list[LocationInformation]:
        pass

    @abstractmethod
    def get_all_locations_collection(self) -> Collection:
        pass

    @abstractmethod
    def get_location(self, location_id: int) -> Location:
        pass

    @abstractmethod
    def get_locations(self, location_ids: list[int]) -> list[Location]:
        pass

    @abstractmethod
    def get_devices(self, location_id: int) -> list[Device]:
        pass

    @abstractmethod
    def load_measurements(
        self,
        components: list[Component],
        date_range: DateRange,
    ) -> None:
        pass

    @abstractmethod
    def get_comfort_profile_setpoints(
        self, date_range: DateRange, location: Location | int
    ) -> list[ComfortProfile]:
        pass

    @abstractmethod
    def get_price_components(
        self,
        supply_point_ids: list[int],
    ) -> dict[int, list[AppliedPriceComponent]]:
        pass

    @abstractmethod
    def get_price_data(
        self,
        date_range: DateRange,
        price_component_ids: list[int],
    ) -> dict[int, PriceData]:
        pass

    @abstractmethod
    def put_setpoint_schedule(
        self, schedule: SetpointSchedule, control_unit: Unit
    ) -> Response:
        pass

    @abstractmethod
    def get_setpoint_schedule(
        self,
        control_unit: Unit,
        date_range: DateRange,
    ) -> SetpointSchedule:
        pass

    @abstractmethod
    def get_electricity_prices(
        self,
        unit: Unit,
        date_range: DateRange,
        include_vat: bool,
        include_tariff: bool,
    ) -> ElectricityPrices:
        pass

    def location_post_setup(self, location: Location) -> None:
        pass

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        for method_name in dir(cls):
            if any(method_name.startswith(s) for s in ["get_", "put_", "load_"]):
                method = getattr(cls, method_name, None)
                if method and not hasattr(method, "_is_logged"):
                    # Apply decorator and mark the method as already decorated
                    logged_method = log_method_call(logger=logger)(method)
                    setattr(logged_method, "_is_logged", True)
                    setattr(cls, method_name, logged_method)


class Factory(ABC):
    _return_class: Callable
    _translation_dict: dict[str, str]

    def __init__(self, input_dict: dict[str, Any] = {}) -> None:
        self.input_dict = input_dict

    def make_sub_classes(self) -> None:
        pass

    @final
    def build(self, **kwargs) -> Any:
        self.input_dict.update(kwargs)
        self.make_sub_classes()
        obj = self._return_class(
            **{
                key: self.input_dict.pop(val)
                for key, val in self._translation_dict.items()
                if val in self.input_dict
            }
        )
        for key, val in self.input_dict.items():
            logger.warning(
                "Unused key-value pair in %s: (%s: %s) Context: %s",
                self.__class__,
                key,
                val,
                obj,
            )
        return obj

    @classmethod
    def translate(self, obj) -> Any:
        if isinstance(obj, self._return_class):
            return {
                val: getattr(obj, key) for key, val in self._translation_dict.items()
            }
        else:
            raise TypeError(
                f"Wrong object type to translate, received {type(obj)} but expected {self._return_class}."
            )



import copy
import glob
import pathlib
import random
import string
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import partial
from typing import ClassVar

import numpy as np
import pandas as pd
import yaml
from requests.models import Response

import preheat_open as po
import preheat_open.api as papi


@dataclass
class MockApiAdapter(po.Adapter):
    building_model: ClassVar[po.BuildingModel] = papi.building_model
    mocks_path: str = str(pathlib.Path(__file__).parent.resolve())

    def open(self):
        pass

    def close(self):
        pass

    def get_location(self, location_id: int) -> po.Location:
        path = f"{self.mocks_path}/locations/{location_id}.yaml"
        with open(path, "r") as file:
            building_model_dict = yaml.safe_load(file)
        return po.Location.from_building_model_dict(
            data=building_model_dict, adapter=self
        )

    def get_locations(self, location_ids: list[int]) -> list[po.Location]:
        return [self.get_location(location_id) for location_id in location_ids]

    def get_all_locations_information(self) -> list[po.LocationInformation]:
        directory_path = (
            pathlib.Path(__file__).parent.resolve() / "test/mocks/locations"
        )
        yaml_files = glob.glob(str(directory_path / "*.yaml"))
        loc_info = []
        for file in yaml_files:
            with open(file, "r") as f:
                building_model_dict = yaml.safe_load(f)
            loc = po.Location.from_building_model_dict(
                data=building_model_dict, adapter=self
            )
            loc_info.append(loc.information)
        return loc_info

    def get_all_locations_collection(self) -> po.Collection:
        return po.Collection.from_location_information_list(
            location_information_list=self.get_all_locations_information(),
            adapter=self,
        )

    def get_devices(self, location_id: int) -> list[po.Device]:
        return []

    def load_measurements(
        self,
        components: list[po.Component],
        date_range: po.DateRange,
        timestamp_type: papi.TimestampType = papi.TimestampType.ISO,
    ) -> None:
        for component in components:
            loadabletype = component.measurements[date_range.resolution]
            mocked_data = mock_component_data(
                date_range=date_range, component_type=component.type
            ).rename(
                component.id if isinstance(component.parent, po.Unit) else component.cid
            )
            loadabletype.update(date_range=date_range, data=mocked_data)

    def get_price_components(
        self,
        supply_point_ids: list[int],
    ) -> dict[int, list[po.AppliedPriceComponent]]:
        pass

    def get_price_data(
        self,
        date_range: po.DateRange,
        price_component_ids: list[int],
        timestamp_type: papi.TimestampType = papi.TimestampType.ISO,
    ) -> dict[int, po.PriceData]:
        pass

    def get_setpoint_schedule(
        self,
        date_range: po.DateRange,
        control_unit: po.Unit,
    ) -> po.SetpointSchedule:
        date_range = copy.deepcopy(date_range)
        date_range.resolution = po.TimeResolution.HOUR
        dr = date_range.to_pandas_date_range()
        values = np.random.uniform(10, 50, len(dr))
        return po.SetpointSchedule(
            schedule=[
                po.ScheduleItem(start=t, value=v, operation=po.OperationType.NORMAL)
                for t, v in zip(dr, values)
            ]
        )

    def put_setpoint_schedule(
        self,
        schedule: po.SetpointSchedule,
        control_unit: po.Unit,
    ) -> Response:
        pass

    def get_comfort_profile_setpoints(self, date_range, location) -> po.ComfortProfiles:
        date_range = copy.deepcopy(date_range)
        comfort_profile_ids = (
            cp
            if (cp := [zone.comfort_profile_id for zone in location.get_zones()])
            else [1]
        )
        date_range.resolution = po.TimeResolution.HOUR
        dr = date_range.to_pandas_date_range()
        return po.ComfortProfiles(
            profiles={
                _id: po.ComfortProfile(
                    id=_id, setpoints=[po.Setpoint(time=t, setpoint=21) for t in dr]
                )
                for _id in comfort_profile_ids
            }
        )

    def get_electricity_prices(
        self, date_range, unit, include_vat, include_tariff
    ) -> po.ElectricityPrices:
        date_range = copy.deepcopy(date_range)
        date_range.resolution = po.TimeResolution.HOUR
        dr = date_range.to_pandas_date_range()
        return po.ElectricityPrices(
            data=[
                po.ElectricityPriceItem(time=t, value=np.random.uniform(0, 4))
                for t in dr
            ],
            tariff_included=include_tariff,
            vat_included=include_vat,
        )

    def get_features(self, location_id):
        pass

    def location_post_setup(self, location: po.Location):
        location.comfort_profiles = (
            po.LoadableData(
                data=po.ComfortProfiles(),
                getter=partial(
                    self.get_comfort_profile_setpoints,
                    location=location,
                ),
            )
            if location.comfort_profiles is None
            else location.comfort_profiles
        )
        for u in location.get_units(papi.UnitQuery(type=papi.UnitType.CONTROL)):
            u.setpoint_schedule = (
                po.LoadableData(
                    data=po.SetpointSchedule(),
                    getter=partial(
                        self.get_setpoint_schedule,
                        control_unit=u,
                    ),
                    setter=partial(
                        self.put_setpoint_schedule,
                        control_unit=u,
                    ),
                )
                if u.setpoint_schedule is None
                else u.setpoint_schedule
            )
        for u in location.get_units(
            papi.UnitQuery(type=[papi.UnitType.ELECTRICITY, papi.UnitType.SUB_METER])
        ):
            if (
                u.type == papi.UnitType.SUB_METER
                and u.parent.type != papi.UnitType.ELECTRICITY
            ):
                continue
            elif u.electricity_prices is None:
                u.electricity_prices = {
                    "tariff_included"
                    if t
                    else "tariff_excluded": {
                        "vat_included"
                        if v
                        else "vat_excluded": po.LoadableData(
                            data=po.ElectricityPrices(),
                            getter=partial(
                                self.get_electricity_prices,
                                unit=u,
                                include_vat=v,
                                include_tariff=t,
                            ),
                        )
                        for v in [True, False]
                    }
                    for t in [True, False]
                }


@dataclass
class ComponentCharacteristic(ABC):
    @abstractmethod
    def mock(self, data: pd.Series) -> pd.Series:
        pass


@dataclass
class CCNoisy(ComponentCharacteristic):
    mean: float = 0.0
    std: float = 1.0

    def mock(self, data: pd.Series) -> pd.Series:
        return pd.Series(
            np.random.normal(self.mean, self.std, len(data)), index=data.index
        )


@dataclass
class CCMonotonicAscending(ComponentCharacteristic):
    def mock(self, data: pd.Series) -> pd.Series:
        return data.abs().cumsum()


@dataclass
class CCRangeConstricted(ComponentCharacteristic):
    min: float = 0.0
    max: float = 100.0

    def mock(self, data: pd.Series) -> pd.Series:
        imin = data.min()
        imax = data.max()
        if imin == imax:
            data = pd.Series(
                np.full(len(data), self.max / 2 - self.min / 2), index=data.index
            )
        else:
            data = self.min + (self.max - self.min) * (data - imin) / (imax - imin)
        return data


@dataclass
class CCYearlyPeriodic(ComponentCharacteristic):
    amplitude: float = 1.0
    day_of_year_top: float = 365 / 2

    def mock(self, data: pd.Series) -> pd.Series:
        return data + pd.Series(
            self.amplitude
            * np.sin(2 * np.pi * (data.index.dayofyear - self.day_of_year_top) / 365),
            index=data.index,
        )


@dataclass
class CCDailyPeriodic(ComponentCharacteristic):
    amplitude: float = 1.0
    hour_of_day_top: float = 12

    def mock(self, data: pd.Series) -> pd.Series:
        return data + pd.Series(
            self.amplitude
            * np.sin(2 * np.pi * (data.index.hour - self.hour_of_day_top) / 24),
            index=data.index,
        )


def get_component_type_mocks(
    component_type: papi.ComponentType,
) -> list[ComponentCharacteristic]:
    lookup = {
        papi.ComponentType.FLOW: [CCNoisy(), CCRangeConstricted(min=0.0, max=500.0)],
        papi.ComponentType.SUPPLY_TEMPERATURE: [
            CCNoisy(),
            CCDailyPeriodic(hour_of_day_top=19),
            CCRangeConstricted(min=10, max=100),
        ],
        papi.ComponentType.RETURN_TEMPERATURE: [
            CCNoisy(),
            CCDailyPeriodic(hour_of_day_top=19),
            CCRangeConstricted(min=10, max=100),
        ],
        papi.ComponentType.PRIMARY_SUPPLY_TEMPERATURE: [
            CCNoisy(),
            CCDailyPeriodic(hour_of_day_top=19),
            CCRangeConstricted(min=10, max=100),
        ],
        papi.ComponentType.PRIMARY_RETURN_TEMPERATURE: [
            CCNoisy(),
            CCDailyPeriodic(hour_of_day_top=19),
            CCRangeConstricted(min=10, max=100),
        ],
        papi.ComponentType.VOLUME: [
            CCNoisy(),
            CCYearlyPeriodic(day_of_year_top=30),
            CCRangeConstricted(min=0, max=20),
            CCMonotonicAscending(),
        ],
        papi.ComponentType.ENERGY: [
            CCNoisy(),
            CCYearlyPeriodic(day_of_year_top=30),
            CCRangeConstricted(min=0, max=20),
            CCMonotonicAscending(),
        ],
        papi.ComponentType.POWER: [
            CCNoisy(),
            CCYearlyPeriodic(day_of_year_top=30),
            CCRangeConstricted(min=0, max=20),
        ],
        papi.ComponentType.HP_ON: [CCNoisy(), CCRangeConstricted(min=0.0, max=1.0)],
        papi.ComponentType.MOTOR_POSITION: [
            CCNoisy(),
            CCRangeConstricted(min=0.0, max=100.0),
        ],
        papi.ComponentType.AMBIENT_TEMPERATURE: [
            CCNoisy(),
            CCDailyPeriodic(),
            CCYearlyPeriodic(day_of_year_top=200),
            CCRangeConstricted(min=-10.0, max=40.0),
        ],
        papi.ComponentType.TEMPERATURE: [
            CCNoisy(),
            CCDailyPeriodic(),
            CCYearlyPeriodic(day_of_year_top=200),
            CCRangeConstricted(min=10, max=40),
        ],
        papi.ComponentType.HUMIDITY: [CCNoisy(), CCRangeConstricted(min=50, max=80)],
        papi.ComponentType.CONTROL_INPUT: [
            CCNoisy(),
            CCRangeConstricted(min=50, max=60),
        ],
        papi.ComponentType.CONTROLLED_SIGNAL: [
            CCNoisy(),
            CCRangeConstricted(min=45, max=65),
        ],
    }
    return lookup[component_type] if component_type in lookup else [CCNoisy()]


def mock_component_data(
    date_range: po.DateRange, component_type: papi.ComponentType
) -> pd.Series:
    index = date_range.to_pandas_date_range()
    series = pd.Series(np.zeros(shape=len(index)), index=index)
    for chts in get_component_type_mocks(component_type):
        series = chts.mock(series)
    return series


def generate_random_string(length=8):
    return "".join(random.choices(string.ascii_letters + string.digits, k=length))


# Helper function to generate unique random integers
def generate_unique_random_ints(count, start=1000, end=9999):
    return random.sample(range(start, end + 1), count)


# Main function to edit the dictionary
def anonymize_building_model_dict(
    building_model_dict: dict, mock_id: int, unit_ids_to_mock: list[int]
) -> tuple[dict, dict[int, int]]:
    random.seed(mock_id)
    mock_unit_mapping = {}
    keys_to_replace = {
        "name": generate_random_string,
        "description": generate_random_string,
        "address": generate_random_string,
        "city": generate_random_string,
        "country": generate_random_string,
        "label": generate_random_string,
        "postcode": generate_random_string,
        "organisation_name": generate_random_string,
        "parent_organisation_name": generate_random_string,
        "latitude": lambda: random.uniform(0, 90),
        "longitude": lambda: random.uniform(0, 180),
        "area": lambda: random.uniform(0, 1000),
        "zipcode": lambda: str(random.randint(1000, 9999)),
    }
    dictionary = {}

    def anonymize(key, d, ids, pkey="information"):
        if isinstance(d, dict):
            return {k: anonymize(k, v, ids, key) for k, v in d.items()}
        elif isinstance(d, list):
            return [anonymize(key, v, ids, pkey) for v in d]
        elif key in keys_to_replace:
            return keys_to_replace[key]()
        elif d is None:
            return None
        elif key.endswith("id") or key.endswith("Id"):
            if key == "unit_id":
                key = "id"
                pkey = "children"
            dkey = f"{pkey}_{key}_{d}"
            if dkey in dictionary:
                return dictionary[dkey]
            dictionary[dkey] = ids.pop()
            if d in unit_ids_to_mock:
                mock_unit_mapping[d] = dictionary[dkey]
            return dictionary[dkey]
        else:
            return d

    # Generate unique random integers for keys ending in "id"
    unique_ints = generate_unique_random_ints(10000, end=99999)

    anon = anonymize("information", building_model_dict, unique_ints)

    def add_consistency(key, d):
        if isinstance(d, dict):
            return {k: add_consistency(k, v) for k, v in d.items()}
        elif isinstance(d, list):
            return [add_consistency(key, v) for v in d]
        elif key == "related_units":
            for key_try in [f"children_id_{d}", f"units_id_{d}"]:
                if key_try in dictionary:
                    return dictionary[key_try]
        elif key in ("zones", "adjacentZones"):
            for key_try in [f"zones_id_{d}", f"subZones_id_{d}"]:
                if key_try in dictionary:
                    return dictionary[key_try]
        else:
            return d

    anon = add_consistency(None, anon)

    anon["id"] = mock_id
    anon["information"]["id"] = mock_id

    return anon, mock_unit_mapping




# This file was generated by 'versioneer.py' (0.20) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2025-06-25T13:33:29+0200",
 "dirty": false,
 "error": null,
 "full-revisionid": "519a7c4f0ef6c7a3d5550c4018a2c01bbb060d3e",
 "version": "2.2"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



"""
This module defines the basics of the interaction with Neogrid's web API
"""
import logging
from enum import Enum
from typing import Any, Optional

from requests import Session
from requests.adapters import HTTPAdapter, Retry
from requests.models import Response

from preheat_open.api.configuration import NeogridApiConfig, run_if_production_mode

from ..time import ZoneInfo, datetime
from .configuration import RuntimeMode

logger = logging.getLogger(__name__)


def api_string_to_datetime(t: str) -> datetime:
    """
    Converts a datetime string from the API to a python datetime

    :param t: datetime in string format
    :return: datetime corresponding to the input
    """
    return datetime.strptime(t, "%Y-%m-%dT%H:%M:%S.%fZ").replace(tzinfo=ZoneInfo("UTC"))


def datetime_to_api_string(t: datetime) -> str:
    """
    Converts a python datetime to a string ready for use in the API I/O

    :param t: datetime to convert to string
    :return: a string with a format compatible with the API
    """
    ts = t.replace(tzinfo=None)
    if (utcoffset := t.utcoffset()) is not None:
        ts = ts - utcoffset
    return ts.strftime("%Y-%m-%dT%H:%M:%SZ")


class APIDataExtractError(Exception):
    pass


def prepare(item: Any, as_json: bool = False) -> Any:
    """
    Method to convert a parameter to a format that can be sent to the API

    :param param: parameter to convert
    :return: converted parameter
    """
    if isinstance(item, str):
        return item
    elif any(isinstance(item, t) for t in [int, float, bool]):
        if as_json:
            return item
        return str(item)
    elif isinstance(item, datetime):
        return datetime_to_api_string(item)
    elif isinstance(item, Enum):
        return item.value
    elif isinstance(item, list):
        if as_json:
            return [prepare(p, as_json=as_json) for p in item]
        return ",".join(map(str, [prepare(p) for p in item]))
    elif isinstance(item, dict):
        if as_json:
            return {k: prepare(v, as_json=as_json) for k, v in item.items()}
        return {k: prepare(v) for k, v in item.items()}
    else:
        raise ValueError(f"Unsupported API input type: {type(item)}")


class ApiSession:
    """Class to handle API connection sessions"""

    def __init__(self, config: Optional[NeogridApiConfig] = None):
        # Adding protection against remote closing connection
        self.config = config if config else NeogridApiConfig()
        retry_strategy = Retry(
            total=3,
            read=3,
            connect=3,
            status_forcelist=(429, 500, 502, 503, 504),
            allowed_methods=["GET", "PUT", "POST"],
            raise_on_status=False,
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        http_session = Session()
        http_session.mount("https://", adapter)
        http_session.mount("http://", adapter)

        self.timeout_seconds = {
            "GET": 120,
            "PUT": 10,
            "POST": 10,
            "PATCH": 10,
            "DELETE": 10,
        }
        self.base_url = self.config.url
        self.session = http_session
        self.session.headers.update({"Authorization": "Apikey " + self.config.token})
        self.datetime_format = "%Y-%m-%dT%H:%M:%S.%f%z"
        logger.debug("Creating API session [%s]...", self)

    def close(self):
        """Method to close the session"""
        self.session.close()
        logger.debug("Closing API session [%s]...", self)

    def __str__(self) -> str:
        """
        Making a string representation of the session object for use in logging

        :return: string representing the session
        """
        return f"ApiSession({hex(id(self.session))})"

    def __log_api_io(self, msg: str) -> None:
        logger.debug(msg)

    def __check_and_raise_error(self, response: Response):
        logger.debug("%s %s", response.status_code, response.reason)

        if not response.ok:
            msg = f"""{response.request.method} - FAILED
            URL: {response.request.url}
            Body: {str(response.request.body)}
            Response: {response.text}
            """
            logger.error(msg)
            response.raise_for_status()

    def __send_request(
        self,
        method: str = "GET",
        endpoint: str = "",
        out: str = "json",
        json_body: dict[str, Any] | None = None,
        params: dict[str, Any] | None = None,
    ) -> Response:
        """
        Method to send a POST request to the API

        :param method: Type of request to send (GET, PUT, POST)
        :param endpoint: URL of the endpoint
        :param out: output type (json or csv)
        :param json: json for the request body
        :param params: params for re request
        :return: API response in request.Response format
        """

        self.__log_api_io(f"{method} /{endpoint} [{self}]")

        path = self.base_url + "/" + endpoint if endpoint else self.base_url

        headers = {"Accept": "text/csv"} if out == "csv" else {}

        if method == "GET":
            response = self.session.request(
                method=method,
                url=path,
                params=prepare(params) if params is not None else params,
                headers=headers,
                json=prepare(json_body, as_json=True)
                if json_body is not None
                else json_body,
                timeout=self.timeout_seconds[method],
            )
        else:
            response = run_if_production_mode(
                function=self.session.request,
                method=method,
                url=path,
                params=prepare(params) if params is not None else params,
                headers=headers,
                json=prepare(json_body, as_json=True)
                if json_body is not None
                else json_body,
                timeout=self.timeout_seconds[method],
            )

        self.__check_and_raise_error(response)

        return response

    def get(
        self,
        endpoint: str = "",
        out: str = "json",
        params: Optional[dict[str, Any]] = None,
    ):
        """
        Method to send a GET request to the API

        :param endpoint: URL of the endpoint
        :param out: output type (json or csv)
        :param params: params for the request
        :return: API response in request.Response format
        """
        return self.__send_request(
            method="GET", endpoint=endpoint, out=out, params=params
        )

    def put(
        self,
        endpoint: str = "",
        out: str = "json",
        json_body: dict[str, Any] | None = None,
        params: dict[str, Any] | None = None,
    ) -> Response:
        """
        Method to send a PUT request to the API

        :param endpoint: URL of the endpoint
        :param out: output type (json or csv)
        :param json: payload for the request
        :param params: params for the request
        :return: API response in request.Response format
        """
        return self.__send_request(
            method="PUT", endpoint=endpoint, out=out, params=params, json_body=json_body
        )

    def post(
        self,
        endpoint: str = "",
        out: str = "json",
        json_body: dict[str, Any] | None = None,
        params: dict[str, Any] | None = None,
    ) -> Response:
        """
        Method to send a POST request to the API

        :param endpoint: URL of the endpoint
        :param out: output type (json or csv)
        :param json_payload: payload for the request
        :param timeout: timeout (in seconds) for the call - defaults to PUT_TIMEOUT_SECONDS
        :return: API response in request.Response format
        """
        return self.__send_request(
            method="POST",
            endpoint=endpoint,
            out=out,
            params=params,
            json_body=json_body,
        )

    def patch(
        self,
        endpoint: str = "",
        out: str = "json",
        json_body: dict[str, Any] | None = None,
        params: dict[str, Any] | None = None,
    ) -> Response:
        """
        Method to send a PATCH request to the API

        :param endpoint: URL of the endpoint
        :param out: output type (json or csv)
        :param json_payload: payload for the request
        :param timeout: timeout (in seconds) for the call - defaults to PUT_TIMEOUT_SECONDS
        :return: API response in request.Response format
        """
        return self.__send_request(
            method="PATCH",
            endpoint=endpoint,
            out=out,
            params=params,
            json_body=json_body,
        )

    def delete(
        self, endpoint: str = "", params: dict[str, Any] | None = None
    ) -> Response:
        """
        Method to send a DELETE request to the API

        :param endpoint: URL of the endpoint
        :param params: params for the request
        :return: API response in request.Response format
        """
        return self.__send_request(method="DELETE", endpoint=endpoint, params=params)



from datetime import datetime

import pytest

from ..time import DateRange, SubDateRange, TimeResolution


@pytest.fixture
def date_ranges():
    dr1 = SubDateRange(datetime(2024, 1, 1), datetime(2024, 1, 5))
    dr2 = SubDateRange(datetime(2024, 1, 10), datetime(2024, 1, 15))
    dr3 = SubDateRange(datetime(2024, 1, 5), datetime(2024, 1, 10))
    dr4 = SubDateRange(datetime(2024, 1, 12), datetime(2024, 1, 20))
    return DateRange([dr1, dr2]), DateRange([dr3, dr4])


def test_add(date_ranges):
    ranges1, ranges2 = date_ranges
    combined = ranges1 + ranges2
    assert len(combined.ranges) == 3


@pytest.mark.parametrize(
    "date, expected",
    [
        (datetime(2024, 1, 3), True),
        (datetime(2024, 1, 8), False),
    ],
)
def test_contains(date_ranges, date, expected):
    ranges1, _ = date_ranges
    assert (date in ranges1) == expected


@pytest.mark.parametrize("expected_len", [1])
def test_intersection(date_ranges, expected_len):
    ranges1, ranges2 = date_ranges
    intersections = ranges1.intersection(ranges2)
    assert len(intersections.ranges) == expected_len


@pytest.mark.parametrize("expected_len", [1])
def test_gaps(date_ranges, expected_len):
    ranges1, _ = date_ranges
    gaps = ranges1.gaps()
    assert len(gaps.ranges) == expected_len


@pytest.mark.parametrize(
    "expected_start, expected_end",
    [
        (datetime(2024, 1, 1), datetime(2024, 1, 15)),
    ],
)
def test_union(date_ranges, expected_start, expected_end):
    ranges1, _ = date_ranges
    union_range = ranges1.union()
    assert union_range.lstart == expected_start
    assert union_range.rend == expected_end


def test_time_resolution():
    res = ["raw", "minute", "hour", "day", "week", "month", "year"]
    for i, ri in enumerate(res):
        for j, rj in enumerate(res):
            assert (i < j) == (TimeResolution(ri) > TimeResolution(rj))

    assert TimeResolution("raw").pandas_alias is None
    assert TimeResolution("minute").pandas_alias == "5T"
    assert TimeResolution("hour").pandas_alias == "H"
    assert TimeResolution("day").pandas_alias == "D"



import logging
import os
import pickle
import shelve
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from functools import wraps
from typing import Any, Optional

from preheat_open.api.configuration import CacheConfig

from ..interfaces import log_method_call
from ..time import datetime, timedelta

logger = logging.getLogger(__name__)


def io_decorator():
    def decorator(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            self.open()
            result = func(self, *args, **kwargs)
            self.close()
            return result

        return wrapper

    return decorator


@dataclass
class Cache(ABC):
    config: CacheConfig = field(default_factory=CacheConfig)

    def __post_init__(self) -> None:
        """
        Post-initialization processing to set up the cache configuration.
        """
        self.directory = (
            os.path.expanduser(self.config.directory) + f"/{self.__class__.__name__}"
        )
        os.makedirs(self.directory, exist_ok=True)

    @abstractmethod
    def get(self, key: str) -> Any:
        pass

    @abstractmethod
    def set(self, key: str, value: Any, expire: int | None = None) -> None:
        pass

    @abstractmethod
    def delete(self, key: str) -> None:
        pass

    @abstractmethod
    def clear(self) -> None:
        pass

    @abstractmethod
    def open(self) -> None:
        pass

    @abstractmethod
    def close(self) -> None:
        pass

    @abstractmethod
    def check_key(self, key: str) -> bool:
        pass

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        methods_to_log = ["clear", "delete", "get", "set", "check_key"]
        for method_name in methods_to_log:
            method = getattr(cls, method_name, None)
            if method:
                decorated_method = log_method_call(logger=logger)(method)
                decorated_method = io_decorator()(decorated_method)
                setattr(cls, method_name, decorated_method)


def easy_cache(cache_attr: str = "_cache"):
    def decorator_cache(func):
        @wraps(func)
        def wrapper_cache(self, *args, **kwargs):
            prefix = f"{self.__class__.__name__}_{func.__name__}"
            key = f"{prefix}_{args}_{kwargs}"
            cache = getattr(self, cache_attr, None)
            if cache and cache.check_key(key):
                return cache.get(key)
            result = func(self, *args, **kwargs)
            if cache:
                cache.set(key, result)
            return result

        return wrapper_cache

    return decorator_cache


def get_cache(config: Optional[CacheConfig] = None) -> Cache | None:
    """
    Factory function to create a cache instance based on the configuration.

    :param config: Cache configuration.
    :return: An instance of a Cache subclass.
    """
    config = config or CacheConfig()
    classes = Cache.__subclasses__()
    cache_class = next((cls for cls in classes if cls.__name__ == config.type), None)
    if cache_class:
        return cache_class(config=config)


@dataclass
class CacheItem:
    key: str
    expires: datetime
    path: Optional[str] = None
    value: Optional[Any] = None


@dataclass
class SimplePickleCache(Cache):
    _registry: dict[str, CacheItem] = field(default_factory=dict)

    def __post_init__(self) -> None:
        super().__post_init__()
        self._registry_path = os.path.join(self.directory, "book.pkl")

    def open(self) -> None:
        # Load the _book mapping from a file if it exists
        if os.path.exists(self._registry_path):
            with open(self._registry_path, "rb") as f:
                self._registry = pickle.load(f)

    def close(self) -> None:
        # Save the _book mapping to a file
        with open(self._registry_path, "wb") as f:
            pickle.dump(self._registry, f)

    def set(self, key: str, value: Any, expire: timedelta | None = None) -> None:
        # Save the value to a file and update the _book mapping
        if expire is None and self.config.time_to_live is not None:
            expire = timedelta(seconds=self.config.time_to_live)
        if self.check_key(key):
            item = self._registry.get(key)
            file_path = item.path
        else:
            file_path = os.path.join(self.directory, f"{uuid.uuid4()}.pkl")
            item = CacheItem(key=key, path=file_path, expires=datetime.now() + expire)
        with open(file_path, "wb") as f:
            pickle.dump(value, f)
        self._registry[key] = item

    def __get(self, key: str) -> Any:
        if self.check_key(key):
            item = self._registry.get(key)
            if item.path and os.path.exists(item.path):
                with open(item.path, "rb") as f:
                    return pickle.load(f)
        return None

    def get(self, key: str) -> Any:
        # Retrieve the value from the file using the _book mapping
        return self.__get(key)

    def delete(self, key: str) -> None:
        # Delete the file and remove the key from the _book mapping
        if key in self._registry:
            item = self._registry.pop(key, None)
            if item.path and os.path.exists(item.path):
                os.remove(item.path)

    def clear(self) -> None:
        # Clear all cache files and the _book mapping
        for item in self._registry.values():
            if os.path.exists(item.path):
                os.remove(item.path)
        self._registry.clear()

    def check_key(self, key: str) -> bool:
        # Check if the key exists in the _book mapping
        item = self._registry.get(key)
        if not item or item.expires < datetime.now():
            self.delete(key)
            return False
        return True


@dataclass
class ShelveCache(Cache):
    def __post_init__(self) -> None:
        """
        Post-initialization processing to set up the cache configuration.
        """
        self._shelf: shelve.Shelf
        super().__post_init__()

    def open(self) -> None:
        """
        Opens the cache.
        """
        self._shelf = shelve.open(self.directory)

    def close(self) -> None:
        """
        Closes the cache.
        """
        self._shelf.close()

    def get(self, key: str) -> Any:
        """
        Gets a value from the cache.

        :param key: The key for the value.
        :type key: str
        :return: The value.
        :rtype: Any
        """
        try:
            item = self._shelf[key]
        except KeyError:
            return None
        return item.value

    def set(self, key: str, value: Any, expire: timedelta | None = None) -> None:
        """
        Sets a value in the cache.

        :param key: The key for the value.
        :type key: str
        :param value: The value to set.
        :type value: Any
        :param expire: The expiration time for the value.
        :type expire: int
        """
        if self.config.time_to_live is not None and expire is None:
            expire = timedelta(self.config.time_to_live)
        self._shelf[key] = CacheItem(
            key=key, value=value, expires=datetime.now() + expire.total_seconds()
        )

    def delete(self, key: str) -> None:
        """
        Deletes a value from the cache.

        :param key: The key for the value.
        :type key: str
        """
        del self._shelf[key]

    def clear(self) -> None:
        """
        Clears the cache.
        """
        self._shelf.clear()

    def check_key(self, key: str) -> bool:
        """
        Checks whether the key is present within the cache
        """
        return key in self._shelf



from __future__ import annotations

import os
from dataclasses import dataclass, field
from enum import Enum
from functools import lru_cache
from typing import ClassVar, Type

import yaml

from ..query import Query
from ..unit import Component, Device, Unit
from ..zone import Zone


class ControlSettingDataType(Enum):
    ENUM = "ENUM"
    INT = "INT"
    DECIMAL = "DECIMAL"
    BOOL = "BOOL"


class ControlSettingUiType(Enum):
    SELECT = "SELECT"
    NUMBER = "NUMBER"
    RANGE = "RANGE"
    CHECKBOX = "CHECKBOX"


class __ControlSettingTypeEnum(Enum):
    def __init__(
        self,
        id: int,
        key: str,
        default_value: int | float | bool,
        min: int | float,
        max: int | float,
        step: int | float,
        type: ControlSettingDataType,
        ui_type: ControlSettingUiType,
        unit,
        show_in_ui,
        description,
        ui_order,
        ui_advanced,
    ):
        self.key = key
        self.id = id
        self.default_value = default_value
        self.min = min
        self.max = max
        self.step = step
        self.type = (
            type
            if isinstance(type, ControlSettingDataType)
            else ControlSettingDataType(type)
        )
        self.ui_type = (
            ui_type
            if isinstance(ui_type, ControlSettingUiType)
            else ControlSettingUiType(ui_type)
        )
        self.unit = unit
        self.show_in_ui = show_in_ui
        self.description = description
        self.ui_order = ui_order
        self.ui_advanced = ui_advanced

    @classmethod
    def from_key(self, key: str) -> __ControlSettingTypeEnum:
        lookup = {setting.key: setting for setting in self.__members__.values()}
        return lookup[key] if key in lookup else None

    @classmethod
    def contains(cls, value: __ControlSettingTypeEnum) -> bool:
        try:
            cls.from_key(value)
        except:
            return False
        return True


@lru_cache(maxsize=1)
def control_setting_types() -> dict:
    dir_path = os.path.dirname(os.path.realpath(__file__))
    with open(
        file=dir_path + "/control_setting_type.yaml", mode="r", encoding="utf8"
    ) as file:
        return yaml.safe_load(file)


ControlSettingType = __ControlSettingTypeEnum(
    "ControlSettingType",
    {setting["key"]: tuple(setting.values()) for setting in control_setting_types()},
)


class RelatedUnitType(Enum):
    COOLING_COIL = "coolingCoilIds"
    HEATING_COIL = "heatingCoilIds"
    HEATING_SECONDARY = "heatingSecondaryId"
    COOLING_SECONDARY = "coolingSecondaryId"
    ELECTRICITY = "electricityId"
    HEATING = "heatingId"
    COOLING = "coolingId"
    MAIN = "mainId"
    HOT_WATER = "hotWaterId"
    COLD_WATER = "coldWaterId"

    @property
    def unittype(self) -> UnitType:
        lookup = {
            RelatedUnitType.COOLING_COIL: UnitType.COOLING_COIL,
            RelatedUnitType.HEATING_SECONDARY: UnitType.SECONDARY,
            RelatedUnitType.HEATING_COIL: UnitType.HEATING_COIL,
            RelatedUnitType.COOLING_SECONDARY: UnitType.SECONDARY,
            RelatedUnitType.ELECTRICITY: UnitType.ELECTRICITY,
            RelatedUnitType.HEATING: UnitType.HEATING,
            RelatedUnitType.COOLING: UnitType.COOLING,
            RelatedUnitType.MAIN: UnitType.MAIN,
            RelatedUnitType.HOT_WATER: UnitType.HOT_WATER,
            RelatedUnitType.COLD_WATER: UnitType.COLD_WATER,
        }
        return lookup[self] if self in lookup else []


class UnitDescriptorType(Enum):
    PANEL_ANGLE = "panelAngle"
    COMPASS_HEADING = "compassHeading"
    PANEL_PEAK_POWER = "panelPeakPower"
    MAX_CONVERTER_PEAK_POWER = "maxConverterPeakPower"
    PRODUCES = "produces"
    BRAND = "brand"
    HEAT_CURVE_OFFSET = "heatCurveOffset"
    HEAT_PUMP_CAPACITY = "heatpumpCapacity"
    BUILDINGS_COVERED = "buildingsCovered"
    PLACEMENT = "placement"
    PLACED_BEFORE_HEAT_RECOVERY = "placedBeforeHeatRecovery"
    PLACEMENT_NUMBER = "placementNumber"
    ACTIVE = "active"
    METER_TYPE = "meterType"
    CONFIGURATION = "configuration"

    @property
    def expected_type(self) -> Type:
        lookup = {
            UnitDescriptorType.PANEL_ANGLE: float,
            UnitDescriptorType.COMPASS_HEADING: float,
            UnitDescriptorType.PANEL_PEAK_POWER: float,
            UnitDescriptorType.MAX_CONVERTER_PEAK_POWER: float,
            UnitDescriptorType.PRODUCES: str,
            UnitDescriptorType.BRAND: str,
            UnitDescriptorType.HEAT_CURVE_OFFSET: float,
            UnitDescriptorType.HEAT_PUMP_CAPACITY: float,
            UnitDescriptorType.BUILDINGS_COVERED: int,
            UnitDescriptorType.PLACEMENT: str,
            UnitDescriptorType.PLACED_BEFORE_HEAT_RECOVERY: bool,
            UnitDescriptorType.PLACEMENT_NUMBER: int,
            UnitDescriptorType.ACTIVE: bool,
            UnitDescriptorType.METER_TYPE: str,
            UnitDescriptorType.CONFIGURATION: dict,
        }
        return lookup[self] if self in lookup else str


class ComponentType(Enum):
    FLOW = "flow"
    SUPPLY_TEMPERATURE = "supplyT"
    RETURN_TEMPERATURE = "returnT"
    VOLUME = "volume"
    VOLUME_CONSUMPTION = "volumeConsumption"
    CONSUMPTION = "consumption"
    ENERGY = "energy"
    FLOW_ACCUMULATED_SUPPLY_TEMPERATURE = "flowAccumulatedSupplyT"
    FLOW_ACCUMULATED_RETURN_TEMPERATURE = "flowAccumulatedReturnT"
    POWER = "power"
    PRIMARY_FLOW = "primaryFlow"
    PRIMARY_RETURN_TEMPERATURE = "primaryReturnT"
    PRIMARY_SUPPLY_TEMPERATURE = "primarySupplyT"
    SENSOR_1 = "sensor1"
    SENSOR_2 = "sensor2"
    SENSOR_3 = "sensor3"
    SENSOR_4 = "sensor4"
    SENSOR_5 = "sensor5"
    SENSOR_6 = "sensor6"
    SENSOR_7 = "sensor7"
    SENSOR_8 = "sensor8"
    SENSOR_9 = "sensor9"
    SENSOR_10 = "sensor10"
    SENSOR_11 = "sensor11"
    SENSOR_12 = "sensor12"
    SENSOR_13 = "sensor13"
    SENSOR_14 = "sensor14"
    SENSOR_15 = "sensor15"
    SENSOR_16 = "sensor16"
    SENSOR_17 = "sensor17"
    SENSOR_18 = "sensor18"
    SENSOR_19 = "sensor19"
    SENSOR_20 = "sensor20"
    BACKUP_ON = "backupOn"
    DEFROST_ON = "defrostOn"
    HOT_GAS = "hotGas"
    HP_ON = "hpOn"
    START_COUNTER = "startCounter"
    TANK_TEMPERATURE = "tankTemperature"
    CONTROL_INPUT = "controlInput"
    CONTROL_SWITCH = "controlSwitch"
    CONTROLLED_SIGNAL = "controlledSignal"
    HEARTBEAT = "heartbeat"
    MOTOR_POSITION = "motorPosition"
    MOTOR_RANGE = "motorRange"
    THERMOSTAT_MOTOR_POSITION_PERCENT = "thermostatMotorPositionPercent"
    ORIGINAL_SETPOINT = "originalSetPoint"
    PERCENTAGE_CONTROLLER = "percentageController"
    PUMP_ON_OFF = "pumpOnOff"
    RELAY_OFF = "relayOff"
    RELAY_ON = "relayOn"
    SECONDARY_CONTROLLED_SIGNAL = "secondaryControlledSignal"
    SETPOINT = "setPoint"
    SETPOINT_OFFSET = "setPointOffset"
    VOLTAGE_CONTROLLER = "voltageController"
    DIFFERENTIAL_PRESSURE = "differentialPressure"
    PRESSURE = "pressure"
    PUMP = "pump"
    RETURN_PRESSURE = "returnPressure"
    SUPPLY_PRESSURE = "supplyPressure"
    VALVE_CURRENT = "valveCurrent"
    VALVE_OPENING = "valveOpening"
    AIR_TEMPERATURE_AFTER = "airTemperatureAfter"
    AIRTEMPERATURE_BEFORE = "airTemperatureBefore"
    CIRCULATION_PUMP = "circulationPump"
    TANK_BOTTOM_TEMPERATURE = "tankBottomT"
    TANK_TOP_TEMPERATURE = "tankTopT"
    CHARGE_TEMPERATURE = "chargeT"
    BATTERY_VOLTAGE = "batteryVoltage"
    CO2 = "co2"
    FLOOR_TEMPERATURE = "floorTemperature"
    FLOOR_TEMPERATURE_MAX = "floorTemperatureMax"
    FLOOR_TEMPERATURE_MIN = "floorTemperatureMin"
    HUMIDITY = "humidity"
    NOISE_LEVEL_AVG = "noiseLevelAvg"
    NOISE_LEVEL_PEAK = "noiseLevelPeak"
    RADON_LONG_TERM = "radonLongTerm"
    RADON_SHORT_TERM = "radonShortTerm"
    SETPOINT_TEMPERATURE = "setPointTemperature"
    SETPOINT_TEMPERATURE_MAX = "setPointTemperatureMax"
    SETPOINT_TEMPERATURE_MIN = "setPointTemperatureMin"
    TEMPERATURE = "temperature"
    THERMOSTAT_MOTOR_POSITION = "thermostatMotorPosition"
    THERMOSTAT_MOTOR_RANGE = "thermostatMotorRange"
    VOC = "voc"
    WINDOW_OPENING = "windowOpening"
    AMBIENT_TEMPERATURE = "ambientTemperature"
    WIND_DIRECTION = "windDirection"
    WIND_SPEED = "windSpeed"
    EXHAUST_AIR_FLOW = "exhaustAirFlow"
    EXHAUST_AIR_PRESSURE = "exhaustAirPressure"
    EXHAUST_AIR_SPEED = "exhaustAirSpeed"
    EXHAUST_AIR_TEMPERATURE = "exhaustAirTemperature"
    EXTRACT_AIR_FLOW = "extractAirFlow"
    EXTRACT_AIR_HUMIDITY = "extractAirHumidity"
    EXTRACT_AIR_PRESSURE = "extractAirPressure"
    EXTRACT_AIR_SPEED = "extractAirSpeed"
    EXTRACT_AIR_TEMPERATURE = "extractAirTemperature"
    EXTRACT_CO2 = "extractCO2"
    EXTRACT_FILTER_DIFFERENTIAL_PRESSURE = "extractFilterDifferentialPressure"
    HEAT_RECOVERY_EXCHANGER_LOADING = "heatRecoveryExchangerLoading"
    HEAT_RECOVERY_RECOVERED_TEMPERATURE = "heatRecoveryRecoveredTemperature"
    INLET_DAMPER_POSITION = "inletDamperPosition"
    INLET_FAN_SPEED = "inletFanSpeed"
    INTAKE_AIR_FLOW = "intakeAirFlow"
    INTAKE_AIR_PRESSURE = "intakeAirPressure"
    INTAKE_AIR_SPEED = "intakeAirSpeed"
    INTAKE_AIR_TEMPERATURE = "intakeAirTemperature"
    INTAKE_FILTER_DIFFERENTIAL_PRESSURE = "intakeFilterDifferentialPressure"
    OUTLET_DAMPER_POSITION = "outletDamperPosition"
    OUTLET_FAN_SPEED = "outletFanSpeed"
    SETPOINT_AIR_FLOW = "setPointAirFlow"
    SETPOINT_CO2 = "setPointCO2"
    SETPOINT_EXTRACT_TEMPERATURE = "setPointExtractTemperature"
    SETPOINT_HEAT_RECOVERY_SUPPLY_TEMPERATURE = "setPointHeatRecoverySupplyTemperature"
    SETPOINT_HUMIDITY = "setPointHumidity"
    SETPOINT_SUPPLY_TEMPERATURE = "setPointSupplyTemperature"
    SUPPLY_AIR_FLOW = "supplyAirFlow"
    SUPPLY_AIR_PRESSURE = "supplyAirPressure"
    SUPPLY_AIR_SPEED = "supplyAirSpeed"
    SUPPLY_AIR_TEMPERATURE = "supplyAirTemperature"
    VENTILATION_ON = "ventilationOn"
    CURRENT_WINDOW_POSITION = "currentWindowPosition"
    ERROR_WINDOW_POSITION = "errorWindowPosition"
    DAMPER_POSITION = "damperPosition"
    LOADING = "loading"
    RECOVERED_TEMPERATURE = "recoveredTemperature"
    EMITTER_TEMPERATURE = "emitterT"
    BATTERY_PERCENTAGE = "batteryPercentage"
    HCA = "hca"
    RADIATOR_TEMPRATURE = "radiatorTempeture"
    TEMPERATURE_ = "Temperature"
    HUMIDITY_ = "Humidity"
    WIND_DIRECTION_ = "WindDirection"
    WIND_SPEED_ = "WindSpeed"
    PRESSURE_ = "Pressure"
    LOW_CLOUDS_ = "LowClouds"
    MEDIUM_CLOUDS_ = "MediumClouds"
    HIGH_CLOUDS_ = "HighClouds"
    FOG_ = "Fog"
    WIND_GUST_ = "WindGust"
    DEW_POINT_TEMPERATURE_ = "DewPointTemperature"
    CLOUDINESS_ = "Cloudiness"
    PRECIPITATION_ = "Precipitation"
    DIRECT_SUN_POWER_ = "DirectSunPower"
    DIFFUSE_SUN_POWER_ = "DiffuseSunPower"
    SUN_ALTITUDE_ = "SunAltitude"
    SUN_AZIMUTH_ = "SunAzimuth"
    DIRECT_SUN_POWER_VERTICAL_ = "DirectSunPowerVertical"
    CHARGING = "charging"
    HOME = "home"
    PLUGGED_IN = "plugged_in"
    RANGE = "range"
    STATE_OF_CHARGE = "stateOfCharge"
    ACTIVE_EXPORT_COUNTER = "active_export_counter"
    ACTIVE_EXPORT_CONSUMPTION = "active_export_consumption"
    REACTIVE_IMPORT = "reactive_import"
    REACTIVE_EXPORT = "reactive_export"
    APPARENT_ENERGY = "apparent_energy"
    APPARENT_POWER = "apparent_power"
    REACTIVE_POWER = "reactive_power"
    FREQUENCY = "frequency"
    CURRENT = "current"
    DMI_TEMPERATURE = "dmiTemperature"
    SG_RELAY_1 = "sgRelay1"
    SG_RELAY_2 = "sgRelay2"

    def __repr__(self) -> str:
        return self.name


class UnitType(Enum):
    CAR_CHARGER = "carChargers"
    COLD_WATER = "coldWater"
    COOLING = "cooling"
    CUSTOM = "custom"
    ELECTRICITY = "electricity"
    HEAT_PUMP = "heatPumps"
    HEATING = "heating"
    HOT_WATER = "hotWater"
    INDOOR_CLIMATE = "indoorClimate"
    LOCAL_WEATHER_STATION = "localWeatherStations"
    MAIN = "main"
    PV = "pvs"
    VENTILATION = "ventilation"
    WINDOW = "window"
    SUB_METER = "subMeters"
    SECONDARY = "secondaries"
    COOLING_COIL = "coolingCoils"
    CONTROL = "controls"
    FLOOR_HEATING_LOOP = "floorHeatingLoops"
    HEATING_COIL = "heatingCoils"
    SUB_HEATING = "subHeating"
    SECONDARY_HEAT_EXCHANGER = "secondaryHeatExchanger"
    SECONDARY_HEAT_EXCHANGER_WITH_TANK = "secondaryHeatExchangerWithTank"
    SECONDARY_TANK = "secondaryTank"
    HEAT_EXCHANGER = "heatExchangers"
    HEAT_RECOVERY = "heatRecovery"
    RADIATOR = "radiators"
    HCA = "hca"
    WEATHER_FORECAST = "weatherForecast"
    THERMOSTAT = "thermostat"
    CAR = "cars"
    SUB_VENTILATION = "subVentilations"

    def __repr__(self) -> str:
        return self.name

    @property
    def components(self) -> list[ComponentType]:
        lookup = {
            UnitType.WEATHER_FORECAST: [
                ComponentType.TEMPERATURE_,
                ComponentType.HUMIDITY_,
                ComponentType.WIND_DIRECTION_,
                ComponentType.WIND_SPEED_,
                ComponentType.PRESSURE_,
                ComponentType.LOW_CLOUDS_,
                ComponentType.MEDIUM_CLOUDS_,
                ComponentType.HIGH_CLOUDS_,
                ComponentType.FOG_,
                ComponentType.WIND_GUST_,
                ComponentType.DEW_POINT_TEMPERATURE_,
                ComponentType.CLOUDINESS_,
                ComponentType.PRECIPITATION_,
                ComponentType.DIRECT_SUN_POWER_,
                ComponentType.DIFFUSE_SUN_POWER_,
                ComponentType.SUN_ALTITUDE_,
                ComponentType.SUN_AZIMUTH_,
                ComponentType.DIRECT_SUN_POWER_VERTICAL_,
            ],
            UnitType.COLD_WATER: [
                ComponentType.FLOW,
                ComponentType.SUPPLY_TEMPERATURE,
                ComponentType.VOLUME,
                ComponentType.VOLUME_CONSUMPTION,
            ],
            UnitType.COOLING: [
                ComponentType.CONSUMPTION,
                ComponentType.ENERGY,
                ComponentType.FLOW_ACCUMULATED_RETURN_TEMPERATURE,
                ComponentType.FLOW_ACCUMULATED_SUPPLY_TEMPERATURE,
                ComponentType.POWER,
                ComponentType.PRIMARY_FLOW,
                ComponentType.PRIMARY_RETURN_TEMPERATURE,
                ComponentType.PRIMARY_SUPPLY_TEMPERATURE,
                ComponentType.VOLUME,
                ComponentType.VOLUME_CONSUMPTION,
            ],
            UnitType.CUSTOM: [
                ComponentType.SENSOR_1,
                ComponentType.SENSOR_2,
                ComponentType.SENSOR_3,
                ComponentType.SENSOR_4,
                ComponentType.SENSOR_5,
                ComponentType.SENSOR_6,
                ComponentType.SENSOR_7,
                ComponentType.SENSOR_8,
                ComponentType.SENSOR_9,
                ComponentType.SENSOR_10,
                ComponentType.SENSOR_11,
                ComponentType.SENSOR_12,
                ComponentType.SENSOR_13,
                ComponentType.SENSOR_14,
                ComponentType.SENSOR_15,
                ComponentType.SENSOR_16,
                ComponentType.SENSOR_17,
                ComponentType.SENSOR_18,
                ComponentType.SENSOR_19,
                ComponentType.SENSOR_20,
            ],
            UnitType.ELECTRICITY: [
                ComponentType.CONSUMPTION,
                ComponentType.ENERGY,
                ComponentType.POWER,
                ComponentType.ACTIVE_EXPORT_CONSUMPTION,
                ComponentType.ACTIVE_EXPORT_COUNTER,
                ComponentType.REACTIVE_IMPORT,
                ComponentType.REACTIVE_EXPORT,
                ComponentType.APPARENT_ENERGY,
                ComponentType.APPARENT_POWER,
                ComponentType.REACTIVE_POWER,
                ComponentType.FREQUENCY,
                ComponentType.CURRENT,
            ],
            UnitType.HEAT_PUMP: [
                ComponentType.BACKUP_ON,
                ComponentType.DEFROST_ON,
                ComponentType.HOT_GAS,
                ComponentType.HP_ON,
                ComponentType.START_COUNTER,
                ComponentType.TANK_TEMPERATURE,
            ],
            UnitType.HEATING: [
                ComponentType.CONSUMPTION,
                ComponentType.ENERGY,
                ComponentType.FLOW_ACCUMULATED_RETURN_TEMPERATURE,
                ComponentType.FLOW_ACCUMULATED_SUPPLY_TEMPERATURE,
                ComponentType.POWER,
                ComponentType.PRIMARY_FLOW,
                ComponentType.PRIMARY_RETURN_TEMPERATURE,
                ComponentType.PRIMARY_SUPPLY_TEMPERATURE,
                ComponentType.VOLUME,
                ComponentType.VOLUME_CONSUMPTION,
            ],
            UnitType.SUB_HEATING: [
                ComponentType.CONSUMPTION,
                ComponentType.ENERGY,
                ComponentType.FLOW_ACCUMULATED_RETURN_TEMPERATURE,
                ComponentType.FLOW_ACCUMULATED_SUPPLY_TEMPERATURE,
                ComponentType.POWER,
                ComponentType.PRIMARY_FLOW,
                ComponentType.PRIMARY_RETURN_TEMPERATURE,
                ComponentType.PRIMARY_SUPPLY_TEMPERATURE,
                ComponentType.VOLUME,
                ComponentType.VOLUME_CONSUMPTION,
            ],
            UnitType.HOT_WATER: [
                ComponentType.CONSUMPTION,
                ComponentType.ENERGY,
                ComponentType.FLOW_ACCUMULATED_RETURN_TEMPERATURE,
                ComponentType.FLOW_ACCUMULATED_SUPPLY_TEMPERATURE,
                ComponentType.POWER,
                ComponentType.PRIMARY_FLOW,
                ComponentType.PRIMARY_RETURN_TEMPERATURE,
                ComponentType.PRIMARY_SUPPLY_TEMPERATURE,
                ComponentType.VOLUME,
                ComponentType.VOLUME_CONSUMPTION,
            ],
            UnitType.INDOOR_CLIMATE: [
                ComponentType.BATTERY_VOLTAGE,
                ComponentType.CO2,
                ComponentType.FLOOR_TEMPERATURE,
                ComponentType.FLOOR_TEMPERATURE_MAX,
                ComponentType.FLOOR_TEMPERATURE_MIN,
                ComponentType.HUMIDITY,
                ComponentType.NOISE_LEVEL_AVG,
                ComponentType.NOISE_LEVEL_PEAK,
                ComponentType.RADON_LONG_TERM,
                ComponentType.RADON_SHORT_TERM,
                ComponentType.SETPOINT_TEMPERATURE,
                ComponentType.SETPOINT_TEMPERATURE_MAX,
                ComponentType.SETPOINT_TEMPERATURE_MIN,
                ComponentType.TEMPERATURE,
                ComponentType.THERMOSTAT_MOTOR_POSITION,
                ComponentType.THERMOSTAT_MOTOR_RANGE,
                ComponentType.VOC,
                ComponentType.WINDOW_OPENING,
            ],
            UnitType.LOCAL_WEATHER_STATION: [
                ComponentType.AMBIENT_TEMPERATURE,
                ComponentType.WIND_DIRECTION,
                ComponentType.WIND_SPEED,
            ],
            UnitType.MAIN: [
                ComponentType.CONSUMPTION,
                ComponentType.ENERGY,
                ComponentType.FLOW,
                ComponentType.FLOW_ACCUMULATED_RETURN_TEMPERATURE,
                ComponentType.FLOW_ACCUMULATED_SUPPLY_TEMPERATURE,
                ComponentType.POWER,
                ComponentType.RETURN_TEMPERATURE,
                ComponentType.SUPPLY_TEMPERATURE,
                ComponentType.VOLUME,
                ComponentType.VOLUME_CONSUMPTION,
            ],
            UnitType.HEAT_RECOVERY: [
                ComponentType.DAMPER_POSITION,
                ComponentType.LOADING,
                ComponentType.RECOVERED_TEMPERATURE,
            ],
            UnitType.VENTILATION: [
                ComponentType.EXHAUST_AIR_FLOW,
                ComponentType.EXHAUST_AIR_PRESSURE,
                ComponentType.EXHAUST_AIR_SPEED,
                ComponentType.EXHAUST_AIR_TEMPERATURE,
                ComponentType.EXTRACT_AIR_FLOW,
                ComponentType.EXTRACT_AIR_HUMIDITY,
                ComponentType.EXTRACT_AIR_PRESSURE,
                ComponentType.EXTRACT_AIR_SPEED,
                ComponentType.EXTRACT_AIR_TEMPERATURE,
                ComponentType.EXTRACT_CO2,
                ComponentType.EXTRACT_FILTER_DIFFERENTIAL_PRESSURE,
                ComponentType.HEAT_RECOVERY_EXCHANGER_LOADING,
                ComponentType.HEAT_RECOVERY_RECOVERED_TEMPERATURE,
                ComponentType.INLET_DAMPER_POSITION,
                ComponentType.INLET_FAN_SPEED,
                ComponentType.INTAKE_AIR_FLOW,
                ComponentType.INTAKE_AIR_PRESSURE,
                ComponentType.INTAKE_AIR_SPEED,
                ComponentType.INTAKE_AIR_TEMPERATURE,
                ComponentType.INTAKE_FILTER_DIFFERENTIAL_PRESSURE,
                ComponentType.OUTLET_DAMPER_POSITION,
                ComponentType.OUTLET_FAN_SPEED,
                ComponentType.SETPOINT_AIR_FLOW,
                ComponentType.SETPOINT_CO2,
                ComponentType.SETPOINT_EXTRACT_TEMPERATURE,
                ComponentType.SETPOINT_HEAT_RECOVERY_SUPPLY_TEMPERATURE,
                ComponentType.SETPOINT_HUMIDITY,
                ComponentType.SETPOINT_SUPPLY_TEMPERATURE,
                ComponentType.SUPPLY_AIR_FLOW,
                ComponentType.SUPPLY_AIR_PRESSURE,
                ComponentType.SUPPLY_AIR_SPEED,
                ComponentType.SUPPLY_AIR_TEMPERATURE,
                ComponentType.VENTILATION_ON,
            ],
            UnitType.SUB_VENTILATION: [
                ComponentType.EXHAUST_AIR_FLOW,
                ComponentType.EXHAUST_AIR_PRESSURE,
                ComponentType.EXHAUST_AIR_SPEED,
                ComponentType.EXHAUST_AIR_TEMPERATURE,
                ComponentType.EXTRACT_AIR_FLOW,
                ComponentType.EXTRACT_AIR_HUMIDITY,
                ComponentType.EXTRACT_AIR_PRESSURE,
                ComponentType.EXTRACT_AIR_SPEED,
                ComponentType.EXTRACT_AIR_TEMPERATURE,
                ComponentType.EXTRACT_CO2,
                ComponentType.EXTRACT_FILTER_DIFFERENTIAL_PRESSURE,
                ComponentType.HEAT_RECOVERY_EXCHANGER_LOADING,
                ComponentType.HEAT_RECOVERY_RECOVERED_TEMPERATURE,
                ComponentType.INLET_DAMPER_POSITION,
                ComponentType.INLET_FAN_SPEED,
                ComponentType.INTAKE_AIR_FLOW,
                ComponentType.INTAKE_AIR_PRESSURE,
                ComponentType.INTAKE_AIR_SPEED,
                ComponentType.INTAKE_AIR_TEMPERATURE,
                ComponentType.INTAKE_FILTER_DIFFERENTIAL_PRESSURE,
                ComponentType.OUTLET_DAMPER_POSITION,
                ComponentType.OUTLET_FAN_SPEED,
                ComponentType.SETPOINT_AIR_FLOW,
                ComponentType.SETPOINT_CO2,
                ComponentType.SETPOINT_EXTRACT_TEMPERATURE,
                ComponentType.SETPOINT_HEAT_RECOVERY_SUPPLY_TEMPERATURE,
                ComponentType.SETPOINT_HUMIDITY,
                ComponentType.SETPOINT_SUPPLY_TEMPERATURE,
                ComponentType.SUPPLY_AIR_FLOW,
                ComponentType.SUPPLY_AIR_PRESSURE,
                ComponentType.SUPPLY_AIR_SPEED,
                ComponentType.SUPPLY_AIR_TEMPERATURE,
                ComponentType.VENTILATION_ON,
            ],
            UnitType.WINDOW: [
                ComponentType.CURRENT_WINDOW_POSITION,
                ComponentType.ERROR_WINDOW_POSITION,
            ],
            UnitType.CONTROL: [
                ComponentType.CONTROL_INPUT,
                ComponentType.CONTROL_SWITCH,
                ComponentType.CONTROLLED_SIGNAL,
                ComponentType.HEARTBEAT,
                ComponentType.MOTOR_POSITION,
                ComponentType.ORIGINAL_SETPOINT,
                ComponentType.PERCENTAGE_CONTROLLER,
                ComponentType.PUMP_ON_OFF,
                ComponentType.RELAY_OFF,
                ComponentType.RELAY_ON,
                ComponentType.SECONDARY_CONTROLLED_SIGNAL,
                ComponentType.SETPOINT,
                ComponentType.SETPOINT_OFFSET,
                ComponentType.VOLTAGE_CONTROLLER,
                ComponentType.SG_RELAY_1,
                ComponentType.SG_RELAY_2,
            ],
            UnitType.COOLING_COIL: [
                ComponentType.AIR_TEMPERATURE_AFTER,
                ComponentType.AIRTEMPERATURE_BEFORE,
                ComponentType.CIRCULATION_PUMP,
                ComponentType.RETURN_TEMPERATURE,
                ComponentType.SUPPLY_TEMPERATURE,
                ComponentType.VALVE_OPENING,
            ],
            UnitType.FLOOR_HEATING_LOOP: [
                ComponentType.CIRCULATION_PUMP,
                ComponentType.RETURN_TEMPERATURE,
                ComponentType.SUPPLY_TEMPERATURE,
                ComponentType.VALVE_CURRENT,
                ComponentType.VALVE_OPENING,
            ],
            UnitType.HEATING_COIL: [
                ComponentType.AIR_TEMPERATURE_AFTER,
                ComponentType.AIRTEMPERATURE_BEFORE,
                ComponentType.CIRCULATION_PUMP,
                ComponentType.RETURN_TEMPERATURE,
                ComponentType.SUPPLY_TEMPERATURE,
                ComponentType.VALVE_OPENING,
            ],
            UnitType.SECONDARY_HEAT_EXCHANGER: [
                ComponentType.RETURN_TEMPERATURE,
                ComponentType.SUPPLY_TEMPERATURE,
            ],
            UnitType.SUB_METER: [
                ComponentType.CONSUMPTION,
                ComponentType.ENERGY,
                ComponentType.FLOW,
                ComponentType.FLOW_ACCUMULATED_RETURN_TEMPERATURE,
                ComponentType.FLOW_ACCUMULATED_SUPPLY_TEMPERATURE,
                ComponentType.POWER,
                ComponentType.RETURN_TEMPERATURE,
                ComponentType.SUPPLY_TEMPERATURE,
                ComponentType.VOLUME,
                ComponentType.VOLUME_CONSUMPTION,
                ComponentType.ACTIVE_EXPORT_CONSUMPTION,
                ComponentType.ACTIVE_EXPORT_COUNTER,
                ComponentType.REACTIVE_IMPORT,
                ComponentType.REACTIVE_EXPORT,
                ComponentType.APPARENT_ENERGY,
                ComponentType.APPARENT_POWER,
                ComponentType.REACTIVE_POWER,
                ComponentType.FREQUENCY,
                ComponentType.CURRENT,
            ],
            UnitType.SECONDARY_HEAT_EXCHANGER_WITH_TANK: [
                ComponentType.CHARGE_TEMPERATURE,
                ComponentType.RETURN_TEMPERATURE,
                ComponentType.SUPPLY_TEMPERATURE,
                ComponentType.TANK_BOTTOM_TEMPERATURE,
                ComponentType.TANK_TOP_TEMPERATURE,
            ],
            UnitType.SECONDARY_TANK: [
                ComponentType.CHARGE_TEMPERATURE,
                ComponentType.RETURN_TEMPERATURE,
                ComponentType.SUPPLY_TEMPERATURE,
                ComponentType.TANK_BOTTOM_TEMPERATURE,
                ComponentType.TANK_TOP_TEMPERATURE,
            ],
            UnitType.HEAT_EXCHANGER: [
                ComponentType.CONSUMPTION,
                ComponentType.ENERGY,
                ComponentType.FLOW,
                ComponentType.FLOW_ACCUMULATED_RETURN_TEMPERATURE,
                ComponentType.FLOW_ACCUMULATED_SUPPLY_TEMPERATURE,
                ComponentType.POWER,
                ComponentType.RETURN_TEMPERATURE,
                ComponentType.SUPPLY_TEMPERATURE,
                ComponentType.VOLUME,
                ComponentType.VOLUME_CONSUMPTION,
            ],
            UnitType.RADIATOR: [
                ComponentType.EMITTER_TEMPERATURE,
                ComponentType.RETURN_TEMPERATURE,
                ComponentType.SUPPLY_TEMPERATURE,
            ],
            UnitType.HCA: [
                ComponentType.AMBIENT_TEMPERATURE,
                ComponentType.BATTERY_PERCENTAGE,
                ComponentType.HCA,
                ComponentType.RADIATOR_TEMPRATURE,
            ],
            UnitType.SECONDARY: [
                ComponentType.CONSUMPTION,
                ComponentType.DIFFERENTIAL_PRESSURE,
                ComponentType.ENERGY,
                ComponentType.FLOW,
                ComponentType.FLOW_ACCUMULATED_RETURN_TEMPERATURE,
                ComponentType.FLOW_ACCUMULATED_SUPPLY_TEMPERATURE,
                ComponentType.POWER,
                ComponentType.PRESSURE,
                ComponentType.PUMP,
                ComponentType.RETURN_PRESSURE,
                ComponentType.RETURN_TEMPERATURE,
                ComponentType.SUPPLY_PRESSURE,
                ComponentType.SUPPLY_TEMPERATURE,
                ComponentType.VOLUME,
                ComponentType.VOLUME_CONSUMPTION,
            ],
            UnitType.THERMOSTAT: [
                ComponentType.TEMPERATURE,
                ComponentType.MOTOR_POSITION,
                ComponentType.MOTOR_RANGE,
                ComponentType.THERMOSTAT_MOTOR_POSITION_PERCENT,
                ComponentType.SETPOINT,
                ComponentType.BATTERY_VOLTAGE,
            ],
            UnitType.CAR: [
                ComponentType.CHARGING,
                ComponentType.HOME,
                ComponentType.PLUGGED_IN,
                ComponentType.RANGE,
                ComponentType.STATE_OF_CHARGE,
                ComponentType.POWER,
            ],
        }
        return lookup[self] if self in lookup else []

    @property
    def children(self) -> list[UnitType]:
        lookup = {
            UnitType.COLD_WATER: [UnitType.SUB_METER],
            UnitType.COOLING: [UnitType.COOLING_COIL, UnitType.SECONDARY],
            UnitType.CUSTOM: [UnitType.CONTROL],
            UnitType.ELECTRICITY: [UnitType.SUB_METER],
            UnitType.SUB_METER: [UnitType.SUB_METER],
            UnitType.HEAT_PUMP: [UnitType.CONTROL],
            UnitType.HEATING: [
                UnitType.FLOOR_HEATING_LOOP,
                UnitType.HEATING_COIL,
                UnitType.SECONDARY,
                UnitType.SUB_HEATING,
            ],
            UnitType.SUB_HEATING: [
                UnitType.FLOOR_HEATING_LOOP,
                UnitType.SECONDARY,
                UnitType.HEATING_COIL,
                UnitType.SUB_HEATING,
            ],
            UnitType.HOT_WATER: [
                UnitType.SECONDARY_HEAT_EXCHANGER,
                UnitType.SECONDARY_HEAT_EXCHANGER_WITH_TANK,
                UnitType.SECONDARY_TANK,
            ],
            UnitType.LOCAL_WEATHER_STATION: [UnitType.CONTROL],
            UnitType.MAIN: [UnitType.HEAT_EXCHANGER],
            UnitType.VENTILATION: [
                UnitType.CONTROL,
                UnitType.HEAT_RECOVERY,
                UnitType.SUB_VENTILATION,
            ],
            UnitType.SUB_VENTILATION: [UnitType.CONTROL, UnitType.HEAT_RECOVERY],
            UnitType.HEAT_RECOVERY: [UnitType.CONTROL],
            UnitType.SECONDARY: [
                UnitType.CONTROL,
                UnitType.FLOOR_HEATING_LOOP,
                UnitType.HEATING_COIL,
                UnitType.RADIATOR,
                UnitType.SECONDARY,
                UnitType.SUB_METER,
                UnitType.COOLING_COIL,
            ],
            UnitType.COOLING_COIL: [UnitType.CONTROL],
            UnitType.FLOOR_HEATING_LOOP: [UnitType.CONTROL],
            UnitType.HEATING_COIL: [UnitType.CONTROL],
            UnitType.SECONDARY_HEAT_EXCHANGER: [UnitType.CONTROL, UnitType.SUB_METER],
            UnitType.SECONDARY_HEAT_EXCHANGER_WITH_TANK: [
                UnitType.CONTROL,
                UnitType.SUB_METER,
            ],
            UnitType.SECONDARY_TANK: [UnitType.CONTROL, UnitType.SUB_METER],
            UnitType.RADIATOR: [UnitType.CONTROL, UnitType.HCA, UnitType.THERMOSTAT],
            UnitType.THERMOSTAT: [UnitType.CONTROL],
        }
        return lookup[self] if self in lookup else []

    @property
    def descriptors(self) -> list[UnitDescriptorType]:
        lookup = {
            UnitType.COLD_WATER: [
                UnitDescriptorType.BUILDINGS_COVERED,
                UnitDescriptorType.PLACEMENT,
            ],
            UnitType.COOLING: [UnitDescriptorType.BUILDINGS_COVERED],
            UnitType.COOLING_COIL: [
                UnitDescriptorType.PLACED_BEFORE_HEAT_RECOVERY,
                UnitDescriptorType.PLACEMENT_NUMBER,
            ],
            UnitType.CONTROL: [UnitDescriptorType.ACTIVE],
            UnitType.ELECTRICITY: [
                UnitDescriptorType.BUILDINGS_COVERED,
                UnitDescriptorType.METER_TYPE,
                UnitDescriptorType.CONFIGURATION,
            ],
            UnitType.SUB_METER: [
                UnitDescriptorType.METER_TYPE,
                UnitDescriptorType.CONFIGURATION,
                UnitDescriptorType.PLACEMENT,
            ],
            UnitType.HEAT_PUMP: [
                UnitDescriptorType.BRAND,
                UnitDescriptorType.HEAT_CURVE_OFFSET,
                UnitDescriptorType.HEAT_PUMP_CAPACITY,
                UnitDescriptorType.PRODUCES,
            ],
            UnitType.HEATING: [UnitDescriptorType.BUILDINGS_COVERED],
            UnitType.SUB_HEATING: [
                UnitDescriptorType.BUILDINGS_COVERED,
            ],
            UnitType.HEATING_COIL: [
                UnitDescriptorType.PLACED_BEFORE_HEAT_RECOVERY,
                UnitDescriptorType.PLACEMENT_NUMBER,
            ],
            UnitType.HOT_WATER: [UnitDescriptorType.BUILDINGS_COVERED],
            UnitType.MAIN: [UnitDescriptorType.BUILDINGS_COVERED],
            UnitType.PV: [
                UnitDescriptorType.COMPASS_HEADING,
                UnitDescriptorType.MAX_CONVERTER_PEAK_POWER,
                UnitDescriptorType.PANEL_ANGLE,
                UnitDescriptorType.PANEL_PEAK_POWER,
            ],
        }
        return lookup[self] if self in lookup else []

    @property
    def related_units(self) -> list[RelatedUnitType]:
        lookup = {
            UnitType.HEAT_PUMP: [
                RelatedUnitType.ELECTRICITY,
                RelatedUnitType.HEATING,
                RelatedUnitType.COOLING,
                RelatedUnitType.HOT_WATER,
                RelatedUnitType.MAIN,
            ],
            UnitType.CAR_CHARGER: [RelatedUnitType.ELECTRICITY],
            UnitType.SECONDARY_HEAT_EXCHANGER: [RelatedUnitType.COLD_WATER],
            UnitType.SECONDARY_HEAT_EXCHANGER_WITH_TANK: [RelatedUnitType.COLD_WATER],
            UnitType.SECONDARY_TANK: [RelatedUnitType.COLD_WATER],
            UnitType.PV: [RelatedUnitType.ELECTRICITY],
            UnitType.VENTILATION: [
                RelatedUnitType.COOLING_COIL,
                RelatedUnitType.ELECTRICITY,
                RelatedUnitType.HEATING_COIL,
                RelatedUnitType.HEATING_SECONDARY,
                RelatedUnitType.COOLING_SECONDARY,
            ],
            UnitType.SUB_VENTILATION: [
                RelatedUnitType.COOLING_COIL,
                RelatedUnitType.ELECTRICITY,
                RelatedUnitType.HEATING_COIL,
                RelatedUnitType.HEATING_SECONDARY,
                RelatedUnitType.COOLING_SECONDARY,
            ],
        }
        return lookup[self] if self in lookup else []

    @property
    def is_top_level(self) -> bool:
        return self in [
            UnitType.CAR,
            UnitType.CAR_CHARGER,
            UnitType.COLD_WATER,
            UnitType.COOLING,
            UnitType.CUSTOM,
            UnitType.ELECTRICITY,
            UnitType.HEAT_PUMP,
            UnitType.HEATING,
            UnitType.HOT_WATER,
            UnitType.INDOOR_CLIMATE,
            UnitType.LOCAL_WEATHER_STATION,
            UnitType.MAIN,
            UnitType.PV,
            UnitType.VENTILATION,
            UnitType.WINDOW,
        ]


class UnitSubType(Enum):
    DIRECT = "DIRECT"
    HOT_WATER_TANK = "HOT_WATER_TANK"
    COUNTER_FLOW_HEAT_EXCHANGER = "COUNTER_FLOW_HEAT_EXCHANGER"
    HEAT_EXCHANGER_WITH_TANK = "HEAT_EXCHANGER_WITH_TANK"
    ROTARY_WHEEL_HEAT_EXCHANGER = "ROTARY_WHEEL_HEAT_EXCHANGER"
    HEAT_EXCHANGER = "HEAT_EXCHANGER"
    MIXING_LOOP = "MIXING_LOOP"
    PRODUCTION = "PRODUCTION"
    CONSUMPTION = "CONSUMPTION"
    CONTROL_HEAT = "CONTROL_HEAT"
    CONTROL_WATER_CIRCULATION = "CONTROL_WATER_CIRCULATION"
    CONTROL_WATER_TANK = "CONTROL_WATER_TANK"
    CONTROL_OTHER_EXTERNAL = "CONTROL_OTHER_EXTERNAL"
    CONTROL_COOLING = "CONTROL_COOLING"
    CONTROL_HEAT_PUMP_ON_OFF = "CONTROL_HEAT_PUMP_ON_OFF"
    CONTROL_TANK_TEMPERATURE_MIN = "CONTROL_TANK_TEMPERATURE_MIN"
    CONTROL_TANK_TEMPERATURE_MAX = "CONTROL_TANK_TEMPERATURE_MAX"
    CONTROL_OUTDOOR_TEMPERATURE = "CONTROL_OUTDOOR_TEMPERATURE"
    CONTROL_HEAT_PUMP_HEAT_CURVE_OFFSET = "CONTROL_HEAT_PUMP_HEAT_CURVE_OFFSET"
    CONTROL_HEAT_PUMP_BLOCKING = "CONTROL_HEAT_PUMP_BLOCKING"
    CONTROL_HEAT_PUMP_BLOCKING_WITH_MAPPING = "CONTROL_HEAT_PUMP_BLOCKING_WITH_MAPPING"
    CONTROL_HEAT_PUMP_BLOCKING_WITH_SLIDER_MAPPING = (
        "CONTROL_HEAT_PUMP_BLOCKING_WITH_SLIDER_MAPPING"
    )
    CONTROL_ZONE = "CONTROL_ZONE"
    BOTH = "BOTH"
    CONTROL_SG_READY = "CONTROL_SG_READY"


@dataclass(eq=False)
class UnitQuery(Query):
    id: list[int] = field(default_factory=list)
    name: list[str] = field(default_factory=list)
    shared: list[bool] = field(default_factory=lambda: [False, None])
    type: list[UnitType] = field(default_factory=list)
    _type_type: ClassVar[type] = UnitType
    subtype: list[UnitSubType] = field(default_factory=list)
    _subtype_type: ClassVar[type] = UnitSubType
    parent: list[UnitQuery] = field(default_factory=list)
    _type_parent: ClassVar[type] = "self"
    exclude: list[UnitQuery] = field(default_factory=list)
    _type_exclude: ClassVar[type] = "self"
    _class: ClassVar[type] = Unit


@dataclass(eq=False)
class ComponentQuery(Query):
    cid: list[int] = field(default_factory=list)
    id: list[int] = field(default_factory=list)
    name: list[str] = field(default_factory=list)
    unit: list[str] = field(default_factory=list)
    type: list[ComponentType] = field(default_factory=list)
    _type_type: ClassVar[type] = ComponentType
    parent: list[UnitQuery] = field(default_factory=list)
    _type_parent: ClassVar[type] = UnitQuery
    exclude: list[ComponentQuery] = field(default_factory=list)
    _type_exclude: ClassVar[type] = "self"
    _class: ClassVar[type] = Component


@dataclass(eq=False)
class DeviceQuery(Query):
    id: list[int] = field(default_factory=list)
    name: list[str] = field(default_factory=list)
    type: list[str] = field(default_factory=list)
    _class: ClassVar[type] = Device


@dataclass(eq=False)
class ZoneQuery(Query):
    id: list[int] = field(default_factory=list)
    name: list[str] = field(default_factory=list)
    type: list[str] = field(default_factory=list)
    external_identifier: list[str] = field(default_factory=list)
    _class: ClassVar[type] = Zone


class BuildingType(Enum):
    """
    Enum class for building types, effectively decoupling from database values
    """

    # For each of these elements, the value must be one of the ones found by querying:
    #       SELECT DISTINCT(building_type) FROM building_characteristics
    SINGLE_FAMILY_HOUSE = "SINGLE_FAMILY_HOME"
    MULTI_FAMILY_HOUSE = "TERRACED_HOUSE"
    APARTMENT_BUILDING = "APARTMENT_BUILDING"
    OFFICE_BUILDING = "OFFICE_BUILDING"
    SHOPPING_CENTER = "SHOPPING_CENTER"
    SCHOOL = "SCHOOL"
    OTHER = "OTHER"
    TEST = "TEST"



from itertools import chain

import pandas as pd

from ..interfaces import Factory
from ..loadable_types import (
    ComfortProfile,
    ComfortProfiles,
    ElectricityPriceItem,
    OperationType,
    ScheduleItem,
    Setpoint,
)
from ..location import Location, LocationCharacteristics, LocationInformation
from ..supplypoint import (
    ApplicationContext,
    ApplicationType,
    AppliedPriceComponent,
    Authority,
    BillingContext,
    FixedPriceData,
    FixedPricePeriod,
    FormatType,
    PriceArea,
    PriceComponent,
    PriceComponentType,
    PriceData,
    SupplyPoint,
)
from ..time import DateRange
from ..unit import (
    Component,
    ControlSetting,
    ControlSettings,
    Device,
    Unit,
    UnitDescriptors,
)
from ..zone import VentilationInfo, Zone
from .session import api_string_to_datetime
from .types import ComponentType, ControlSettingType, UnitSubType, UnitType


class UnitComponentFactory(Factory):
    _return_class = Component
    _translation_dict = {
        "cid": "cid",
        "id": "id",
        "name": "name",
        "type": "type",
        "unit": "unit",
        "std_unit_divisor": "stdUnitDivisor",
        "std_unit": "stdUnit",
        "min": "min",
        "max": "max",
    }


class ControlSettingFactory(Factory):
    _return_class = ControlSetting
    _translation_dict = {
        "type": "type",
        "value": "value",
        "unit": "unit",
    }

    def make_sub_classes(self) -> None:
        self.input_dict["type"] = ControlSettingType.from_key(
            self.input_dict.pop("key")
        )


class ControlSettingsFactory(Factory):
    _return_class = ControlSettings
    _translation_dict = {
        "settings": "settings",
        "parent": "parent",
    }

    def make_sub_classes(self) -> None:
        parent = self.input_dict.pop("parent")
        self.input_dict = {
            "settings": {
                key: ControlSettingFactory(item | {"key": key}).build()
                for key, item in self.input_dict["settings"].items()
            }
        }
        self.input_dict["parent"] = parent


class UnitFactory(Factory):
    _return_class = Unit
    _translation_dict = {
        "type": "unit_type",
        "subtype": "type",
        "id": "id",
        "name": "name",
        "components": "components",
        "descriptors": "descriptors",
        "zones": "zoneIds",
        "shared": "shared",
        "shared_from": "sharedFrom",
        "shared_locations": "sharedLocations",
        "children": "children",
        "related_units": "related_units",
        "buildings_covered": "buildingsCovered",
        "covers_building": "coversBuilding",
        "control_settings": "control_settings",
        "electricity_prices": "electricity_prices",
        "setpoint_schedule": "setpoint_schedule",
    }

    def make_sub_classes(self) -> None:
        unit_type: UnitType = self.input_dict["unit_type"]
        if "type" in self.input_dict:
            self.input_dict["type"] = (
                UnitSubType(t) if (t := self.input_dict["type"]) else None
            )
        self.input_dict["children"] = [
            UnitFactory(input_dict=child | {"unit_type": child_type}).build()
            for child_type in unit_type.children
            if (
                children := val
                if isinstance((val := self.input_dict.pop(child_type.value, [])), list)
                else [val]
            )
            for child in children
            if child is not None
        ]
        self.input_dict["components"] = [
            UnitComponentFactory(comp | {"type": comp_type}).build()
            for comp_type in unit_type.components
            if (comp := self.input_dict.pop(comp_type.value, None)) is not None
        ]
        self.input_dict["descriptors"] = UnitDescriptors(
            **{
                d.value: self.input_dict.pop(d.value)
                for d in unit_type.descriptors
                if d.value in self.input_dict
            }
        )
        self.input_dict["related_units"] = list(
            chain(
                *[
                    record if isinstance(record, list) else [record]
                    for related_unit in unit_type.related_units
                    if (record := self.input_dict.pop(related_unit.value))
                ]
            )
        )
        if "zoneId" in self.input_dict:
            self.input_dict["zoneIds"] = [self.input_dict.pop("zoneId")]
        if "controlSettings" in self.input_dict:
            self.input_dict["control_settings"] = ControlSettingsFactory(
                {
                    "settings": self.input_dict.pop("controlSettings"),
                    "parent": self.input_dict["id"],
                }
            ).build()


class WeatherComponentFactory(Factory):
    _return_class = Component
    _translation_dict = {
        "id": "id",
        "name": "name",
        "unit": "unit",
        "type": "type",
    }


class WeatherFactory(Factory):
    _return_class = Unit
    _translation_dict = {
        "id": "gridId",
        "name": "name",
        "type": "unit_type",
        "components": "types",
    }

    def make_sub_classes(self) -> None:
        self.input_dict["types"] = [
            WeatherComponentFactory(
                comp | {"type": ComponentType(comp["name"])}
            ).build()
            for comp in self.input_dict["types"]
        ]
        self.input_dict["name"] = "Weather Forecast"


class ZoneFactory(Factory):
    _return_class = Zone
    _translation_dict = {
        "id": "id",
        "name": "name",
        "external_identifier": "externalIdentifier",
        "area": "zoneArea",
        "has_external_wall": "hasExternalWall",
        "sub_zones": "subZones",
        "adjacent_zones": "adjacentZones",
        "type": "type",
        "comfort_profile_id": "comfortProfileId",
        "ventilation_information": "ventilation_information",
    }

    def make_sub_classes(self) -> None:
        self.input_dict["adjacentZones"] = [
            d["zoneId"] for d in self.input_dict["adjacentZones"]
        ]
        self.input_dict["subZones"] = [
            ZoneFactory(d) for d in self.input_dict["subZones"]
        ]
        self.input_dict["ventilation_information"] = VentilationInfo(
            has_ventilation_supply=self.input_dict.pop("ventilationSupply"),
            has_ventilation_exhaust=self.input_dict.pop("ventilationExhaust"),
        )


class LocationCharacteristicsFactory(Factory):
    _return_class = LocationCharacteristics
    _translation_dict = {
        "area": "buildingArea",
        "type": "buildingType",
        "number_of_apartments": "apartments",
    }


class LocationInformationFactory(Factory):
    _return_class = LocationInformation
    _translation_dict = {
        "address": "address",
        "city": "city",
        "country": "country",
        "label": "label",
        "latitude": "latitude",
        "longitude": "longitude",
        "id": "locationId",
        "organisation_id": "organizationId",
        "organisation_name": "organizationName",
        "parent_organisation_id": "parentOrganizationId",
        "parent_organisation_name": "parentOrganizationName",
        "timezone": "timezone",
        "zipcode": "zipcode",
    }


class SupplyPointFactory(Factory):
    _return_class = SupplyPoint
    _translation_dict = {
        "id": "id",
        "name": "name",
        "sub_type": "subType",
        "type": "type",
        "units": "unitIds",
    }


class LocationFactory(Factory):
    _return_class = Location
    _translation_dict = {
        "id": "id",
        "information": "location",
        "zones": "zones",
        "units": "units",
        "characteristics": "characteristics",
        "supply_points": "supply_points",
        "adapter": "adapter",
    }

    def make_sub_classes(self) -> None:
        self.input_dict["units"] = [
            UnitFactory(input_dict=d | {"unit_type": utype}).build()
            for utype in UnitType
            if utype.is_top_level
            for d in self.input_dict.pop(utype.value)
        ] + [
            WeatherFactory(
                input_dict=self.input_dict.pop(UnitType.WEATHER_FORECAST.value)
                | {
                    "unit_type": UnitType.WEATHER_FORECAST,
                }
            ).build()
        ]

        self.input_dict["zones"] = [
            ZoneFactory(zone).build() for zone in self.input_dict["zones"]
        ]
        self.input_dict["location"] = LocationInformationFactory(
            self.input_dict["location"]
        ).build()
        self.input_dict["id"] = self.input_dict["location"].id
        self.input_dict["characteristics"] = LocationCharacteristicsFactory(
            {
                key: self.input_dict.pop(key)
                for key in ["buildingArea", "buildingType", "apartments"]
            }
        ).build()
        self.input_dict["supply_points"] = [
            SupplyPointFactory(sp_i).build()
            for sp_i in self.input_dict.pop("supplyPoints")
        ]


class DeviceComponentFactory(Factory):
    _return_class = Component
    _translation_dict = {
        "cid": "cid",
        "id": "id",
        "max": "max",
        "min": "min",
        "name": "name",
        "type": "typeName",
        "unit": "unit",
    }


class DeviceFactory(Factory):
    _return_class = Device
    _translation_dict = {
        "name": "name",
        "id": "id",
        "type": "typeName",
        "serial": "serial",
        "components": "components",
    }

    def make_sub_classes(self) -> None:
        self.input_dict["components"] = [
            DeviceComponentFactory(r).build() for r in self.input_dict["components"]
        ]


class PriceComponentTypeFactory(Factory):
    _return_class = PriceComponentType
    _translation_dict = {
        "id": "id",
        "name": "Name",
    }


class PriceAreaFactory(Factory):
    _return_class = PriceArea
    _translation_dict = {
        "id": "id",
        "name": "name",
        "supply_type": "supplyType",
    }


class AuthorityFactory(Factory):
    _return_class = Authority
    _translation_dict = {
        "id": "id",
        "name": "name",
    }


class PriceComponentFactory(Factory):
    _return_class = PriceComponent
    _translation_dict = {
        "id": "id",
        "name": "name",
        "description": "description",
        "created": "created",
        "type": "type",
        "unit": "unit",
        "format_type": "formatType",
        "application_type": "applicationType",
        "unit_component_type": "unitComponentType",
        "timezone": "timezone",
        "price_area": "priceArea",
        "authority": "authority",
    }

    def make_sub_classes(self) -> None:
        self.input_dict["type"] = PriceComponentTypeFactory(
            self.input_dict["type"]
        ).build()
        self.input_dict["formatType"] = FormatType(self.input_dict["formatType"])
        self.input_dict["applicationType"] = ApplicationType(
            self.input_dict["applicationType"]
        )
        self.input_dict["priceArea"] = PriceAreaFactory(
            self.input_dict["priceArea"]
        ).build()
        self.input_dict["authority"] = AuthorityFactory(
            self.input_dict["authority"]
        ).build()
        self.input_dict["created"] = api_string_to_datetime(self.input_dict["created"])


class AppliedPriceComponentFactory(Factory):
    _return_class = AppliedPriceComponent
    _translation_dict = {
        "id": "id",
        "supplypoint_id": "supplyPointId",
        "price_component": "priceComponent",
        "contract_id": "contractId",
        "created": "created",
        "valid_range": "valid_range",
        "billing_context": "billingContext",
        "application_context": "applicationContext",
    }

    def make_sub_classes(self) -> None:
        self.input_dict["valid_range"] = DateRange(
            start=api_string_to_datetime(val)
            if (val := self.input_dict.pop("validFrom"))
            else None,
            end=api_string_to_datetime(val)
            if (val := self.input_dict.pop("validTo"))
            else None,
        )
        self.input_dict["priceComponent"] = PriceComponentFactory(
            self.input_dict["priceComponent"]
        ).build()
        self.input_dict["billingContext"] = BillingContext(
            self.input_dict["billingContext"]
        )
        self.input_dict["applicationContext"] = ApplicationContext(
            self.input_dict["applicationContext"]
        )


class FixedPriceDataFactory(Factory):
    _return_class = FixedPriceData
    _translation_dict = {
        "position": "position",
        "price": "price",
    }


class FixedPricePeriodFactory(Factory):
    _return_class = FixedPricePeriod
    _translation_dict = {
        "data": "data",
        "date_range": "dateRange",
        "resolution": "resolution",
    }

    def make_sub_classes(self) -> None:
        self.input_dict["data"] = [
            FixedPriceDataFactory(elem).build() for elem in self.input_dict["data"]
        ]
        self.input_dict["dateRange"] = DateRange(
            start=api_string_to_datetime(val)
            if (val := self.input_dict.pop("validFrom"))
            else None,
            end=api_string_to_datetime(val)
            if (val := self.input_dict.pop("validTo"))
            else None,
        )


class PriceDataFactory(Factory):
    _return_class = PriceData
    _translation_dict = {
        "fixed_price_data": "fixedPriceData",
        "time_series_data": "timeSeriesData",
    }

    def make_sub_classes(self) -> None:
        fpd = (
            []
            if self.input_dict["fixedPriceData"] is None
            else self.input_dict["fixedPriceData"]
        )
        self.input_dict["fixedPriceData"] = [
            FixedPricePeriodFactory(elem).build() for elem in fpd
        ]
        self.input_dict["timeSeriesData"] = (
            pd.Series(data) if (data := self.input_dict["timeSeriesData"]) else None
        )


class ScheduleItemFactory(Factory):
    _return_class = ScheduleItem
    _translation_dict = {
        "start": "startTime",
        "value": "value",
        "owner_id": "ownerId",
        "operation": "operation",
    }

    def make_sub_classes(self) -> None:
        self.input_dict["startTime"] = api_string_to_datetime(
            self.input_dict["startTime"]
        )
        self.input_dict["operation"] = OperationType(self.input_dict["operation"])


class SetpointFactory(Factory):
    _return_class = Setpoint
    _translation_dict = {
        "time": "time",
        "setpoint": "setpoint",
        "vacation": "vacation",
        "min": "min",
        "max": "max",
    }

    def make_sub_classes(self) -> None:
        self.input_dict["time"] = api_string_to_datetime(self.input_dict["time"])


class ComfortProfileFactory(Factory):
    _return_class = ComfortProfile
    _translation_dict = {
        "id": "id",
        "name": "name",
        "setpoints": "setpoints",
    }

    def make_sub_classes(self) -> None:
        self.input_dict["setpoints"] = [
            SetpointFactory(elem).build() for elem in self.input_dict["setpoints"]
        ]


class ComfortProfilesFactory(Factory):
    _return_class = ComfortProfiles
    _translation_dict = {
        "profiles": "comfortProfiles",
    }

    def make_sub_classes(self) -> None:
        self.input_dict["comfortProfiles"] = {
            elem["id"]: ComfortProfileFactory(elem).build()
            for elem in self.input_dict["comfortProfiles"]
        }


class ElectricityPriceItemFactory(Factory):
    _return_class = ElectricityPriceItem
    _translation_dict = {
        "time": "time",
        "value": "value",
    }

    def make_sub_classes(self) -> None:
        self.input_dict["time"] = api_string_to_datetime(self.input_dict["time"])



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime
from typing import ClassVar, Generator

import pytest

from preheat_open.query import (
    NoElementError,
    NoUniqueElementError,
    Query,
    query_stuff,
    unique,
)


@dataclass
class TClass:
    id: int
    attr1: str = ""
    attr2: str = ""
    attr3: list[TClass] = field(default_factory=list)
    attr4: TClass = None


@dataclass(eq=False)
class TClassQuery(Query):
    id: list[int] = field(default_factory=list)
    attr1: list[str] = field(default_factory=list)
    attr2: list[str] = field(default_factory=list)
    attr4: list[TClass] = field(default_factory=list)
    _type_attr4: ClassVar[type] = TClass
    _class: ClassVar[type] = TClass


@pytest.fixture(scope="module")
def test_object0():
    return TClass(id=1, attr1="value1", attr2="value2", attr3=[])


@pytest.fixture(scope="module")
def test_object1():
    return TClass(
        id=2,
        attr1="value1",
        attr2="value3",
        attr3=[
            TClass(id=5, attr1="value1", attr2="value2", attr3=[]),
            TClass(id=4, attr1="value2", attr2="value3", attr3=[]),
        ],
    )


@pytest.fixture(scope="module")
def test_object2():
    return TClass(id=3, attr1="value3", attr2="value2", attr3=[])


@pytest.fixture(scope="module")
def test_query0():
    return TClassQuery(attr1=["value1"])


@pytest.fixture(scope="module")
def test_query1():
    return TClassQuery(id=[2], attr1=["value1"], attr2=["value3"])


@pytest.fixture(scope="module")
def test_query2():
    return TClassQuery(id=[3], attr1=["value3"], attr2=["value2"])


@pytest.mark.parametrize(
    "query, expected",
    [
        (TClassQuery(attr1="value1"), (["value1"], [])),
        (TClassQuery(attr2="value2"), ([], ["value2"])),
    ],
)
def test_query_post_init(query, expected):
    attr1, attr2 = expected
    assert query.attr1 == attr1
    assert query.attr2 == attr2


@pytest.mark.parametrize(
    "obj, query, expected",
    [
        ("test_object0", "test_query0", True),
        ("test_object1", "test_query1", True),
        ("test_object2", "test_query2", True),
        ("test_object0", "test_query1", False),
        ("test_object1", "test_query0", True),
        ("test_object0", "test_query2", False),
        ("test_object2", "test_query0", False),
        ("test_object1", "test_query2", False),
        ("test_object2", "test_query1", False),
    ],
)
def test_query_equal(obj, query, expected, request):
    fobj = request.getfixturevalue(obj)
    fquery = request.getfixturevalue(query)
    assert (fobj == fquery) == expected


def test_query_convert_attr():
    converted = TClassQuery.convert_attr({"id": 0, "attr1": "test"}, "attr4")
    assert isinstance(converted, TClass)
    assert converted.attr1 == "test"


def test_query_from_kwargs():
    kwargs = dict(attr1="value1", attr2="value2")
    query = TClassQuery.from_kwargs(**kwargs)
    assert query.attr1 == ["value1"]
    assert query.attr2 == ["value2"]


@pytest.mark.parametrize(
    "query, obj, expected",
    [
        ("test_query0", "test_object0", 1),
        ("test_query1", "test_object1", 1),
        ("test_query2", "test_object2", 1),
        ("test_query1", "test_object0", 0),
        ("test_query0", "test_object1", 2),
        ("test_query2", "test_object0", 0),
        ("test_query0", "test_object2", 0),
        ("test_query1", "test_object2", 0),
        ("test_query2", "test_object1", 0),
    ],
)
def test_query_stuff(query, obj, expected, request):
    fquery = request.getfixturevalue(query)
    fobj = request.getfixturevalue(obj)
    results = list(
        query_stuff(
            obj=fobj,
            query=fquery,
            query_type=TClassQuery,
            sub_obj_attrs=["attr3"],
            include_obj=True,
        )
    )
    assert len(results) == expected


def test_unique():
    def generator() -> Generator[int, None, None]:
        yield 1

    assert unique(generator()) == 1

    def generator_multiple() -> Generator[int, None, None]:
        yield 1
        yield 2

    with pytest.raises(NoUniqueElementError):
        unique(generator_multiple())

    def generator_none() -> Generator[int, None, None]:
        if False:
            yield

    with pytest.raises(NoElementError):
        unique(generator_none())



import pathlib

import pytest

import preheat_open as po
from preheat_open.api.conftest import mock_adapter
from preheat_open.test import (
    API_KEY,
    MEDIUM_TEST_PERIOD,
    MOCK_LOCATION_ID,
    SHORT_TEST_PERIOD,
    TEST_LOCATION_ID,
)

# # Defining fixture scope
# FIXTURE_SCOPE = "session"

# @pytest.fixture(scope=FIXTURE_SCOPE)
# def configuration():
#     return papi.Configuration.from_file("yaml")

# @pytest.fixture(scope=FIXTURE_SCOPE)
# def medium_period():
#     return MEDIUM_TEST_PERIOD


# @pytest.fixture(scope=FIXTURE_SCOPE)
# def short_period():
#     return SHORT_TEST_PERIOD


# @pytest.fixture(scope=FIXTURE_SCOPE)
# def location():
#     return po.Location(TEST_LOCATION_ID)


# @pytest.fixture(scope=FIXTURE_SCOPE)
# def location_with_devices(location):
#     b = deepcopy(location)
#     b.load_devices()
#     return b


# @pytest.fixture(scope=FIXTURE_SCOPE)
# def location_with_data(location, medium_period):
#     location_new = deepcopy(location)
#     location_new.get_measurements(*medium_period)
#     return location_new


# # Legacy fixtures
# @pytest.fixture(scope=FIXTURE_SCOPE)
# def building_id():
#     return TEST_LOCATION_ID


# @pytest.fixture(scope=FIXTURE_SCOPE)
# def unit_id():
#     return 15312


# @pytest.fixture(scope=FIXTURE_SCOPE)
# def control_unit_id():
#     return 15357


# @pytest.fixture(scope=FIXTURE_SCOPE)
# def unit(location, unit_id):
#     return location.query_units(unit_id=unit_id)[0]


# @pytest.fixture(scope=FIXTURE_SCOPE)
# def unit_with_data(unit, medium_period):
#     unit_new = deepcopy(unit)
#     unit_new.get_measurements(*medium_period)
#     return unit_new


# @pytest.fixture(scope=FIXTURE_SCOPE)
# def control_unit(location):
#     return location.query_units("control", "control_unit_custom_1")[0]


# @pytest.fixture(scope=FIXTURE_SCOPE)
# def weather_unit(location):
#     return location.weather


@pytest.fixture(scope="session")
def location(mock_adapter) -> po.Location:
    return mock_adapter.get_location(MOCK_LOCATION_ID)



import pytest

from preheat_open.zone import VentilationInfo, Zone


@pytest.fixture
def zone():
    sub_zone1 = Zone(
        id=2,
        name="SubZone",
        type="Residential",
        external_identifier="SubZone1",
        area=50.0,
        has_external_wall=True,
        parent=None,
        sub_zones=[],
        units=[],
        adjacent_zones=[],
        comfort_profile_id=11,
        ventilation_information=VentilationInfo(
            has_ventilation_supply=True, has_ventilation_exhaust=False
        ),
    )
    sub_zone2 = Zone(
        id=3,
        name="SubZone2",
        type="Residential",
        external_identifier="SubZone2",
        area=50.0,
        has_external_wall=True,
        parent=None,
        sub_zones=[],
        units=[],
        adjacent_zones=[],
        comfort_profile_id=12,
        ventilation_information=VentilationInfo(
            has_ventilation_supply=True, has_ventilation_exhaust=False
        ),
    )
    return Zone(
        id=1,
        name="MainZone",
        type="Residential",
        external_identifier="Zone1",
        area=100.0,
        has_external_wall=True,
        parent=None,
        sub_zones=[sub_zone1, sub_zone2],
        units=[],
        adjacent_zones=[2, 3],
        comfort_profile_id=10,
        ventilation_information=VentilationInfo(
            has_ventilation_supply=True, has_ventilation_exhaust=False
        ),
    )


def test_zone_post_init(zone):
    assert len(zone.sub_zones) == 2
    assert isinstance(zone.sub_zones[0], Zone)
    assert zone.sub_zones[0].name == "SubZone"


def test_zone_to_building_model_dict(zone):
    expected_dict = {
        "id": 1,
        "name": "MainZone",
        "type": "Residential",
        "externalIdentifier": "Zone1",
        "area": 100.0,
        "hasExternalWall": True,
        "subZones": [z.to_building_model_dict() for z in zone.sub_zones],
        "adjacentZones": [2, 3],
        "comfortProfileId": 10,
        "ventilation_information": {
            "has_ventilation_supply": True,
            "has_ventilation_exhaust": False,
        },
    }
    assert zone.to_building_model_dict() == expected_dict


def test_zone_from_building_model_dict():
    data = {
        "id": 1,
        "name": "MainZone",
        "type": "Residential",
        "externalIdentifier": "Zone1",
        "area": 100.0,
        "hasExternalWall": True,
        "parent": None,
        "subZones": [],
        "units": [],
        "adjacentZones": [2, 3],
        "comfortProfileId": 10,
        "ventilation_information": {
            "has_ventilation_supply": True,
            "has_ventilation_exhaust": False,
        },
    }
    zone = Zone.from_building_model_dict(data)
    assert zone.id == 1
    assert zone.name == "MainZone"
    assert zone.type == "Residential"
    assert zone.external_identifier == "Zone1"
    assert zone.area == 100.0
    assert zone.has_external_wall is True
    assert zone.parent is None
    assert zone.adjacent_zones == [2, 3]
    assert zone.comfort_profile_id == 10
    assert zone.ventilation_information.has_ventilation_supply is True
    assert zone.ventilation_information.has_ventilation_exhaust is False


def test_zone_repr(zone):
    assert repr(zone) == "Zone(type=Residential, name=MainZone, id=1)"
